# Generated by Django 2.1.2 on 2018-10-31 21:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_verifier', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='dataverificationtag',
            options={'permissions': [('can_verify', 'Can verify Data Tag?')]},
        ),
        migrations.AlterField(
            model_name='dataverificationtag',
            name='note',
            field=models.TextField(blank=True, default=''),
        ),
    ]



__all__ = ["FileForm", "TextForm", "RawTextForm"]

from django.db import models
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django import forms

from spkcspider.apps.spider.helpers import get_settings_func
from spkcspider.apps.spider.fields import SanitizedHtmlField
from .models import FileFilet, TextFilet
from .conf import (
    DEFAULT_LICENSE_FILE, DEFAULT_LICENSE_TEXT, LICENSE_CHOICES
)
from .widgets import LicenseChooserWidget

from spkcspider.apps.spider.fields import (
    OpenChoiceField, MultipleOpenChoiceField
)
from spkcspider.apps.spider.widgets import ListWidget, Select2Widget

_extra = '' if settings.DEBUG else '.min'


def check_attrs_func(tag, name, value):
    # currently no objections
    return True


def _extract_choice(item):
    return (item[0], item[1].get("name", item[0]))


class FileForm(forms.ModelForm):
    license_name = OpenChoiceField(
        label=_("License"), help_text=_("Select license"),
        choices=map(_extract_choice, LICENSE_CHOICES.items()),
        widget=LicenseChooserWidget(
            licenses=LICENSE_CHOICES
        )
    )
    sources = MultipleOpenChoiceField(
        required=False, initial=False,
        widget=ListWidget(
            item_label=_("Source")
        )
    )

    class Meta:
        model = FileFilet
        fields = ['file', 'license_name', 'license_url', 'sources']
        widgets = {
            "license_url": forms.URLInput(
                attrs={
                    "style": "width:100%"
                }
            )
        }

    class Media:
        js = [
            'spider_filets/licensechooser.js'
        ]

    def __init__(self, request, uc=None, initial=None, **kwargs):
        if initial is None:
            initial = {}
        if not getattr(kwargs.get("instance", None), "id", None):
            initial.setdefault(
                "license_name", DEFAULT_LICENSE_FILE(uc, request.user)
            )
        super().__init__(initial=initial, **kwargs)
        setattr(self.fields['file'], "hashable", True)
        # sources should not be hashed as they don't affect result
        setattr(self.fields['sources'], "hashable", False)
        setattr(self.fields['license_name'], "hashable", True)
        setattr(self.fields['license_url'], "hashable", True)
        if request.user.is_superuser:
            # no upload limit
            pass
        elif request.user.is_staff:
            self.fields["file"].max_length = getattr(
                settings, "MAX_FILE_SIZE_STAFF", None
            )
        else:
            self.fields["file"].max_length = getattr(
                settings, "MAX_FILE_SIZE", None
            )
        if request.is_owner:
            # self.user = request.user
            return
        self.fields["file"].editable = False
        self.fields["name"].editable = False

    def clean(self):
        ret = super().clean()
        if "file" not in ret:
            return ret
        # has to raise ValidationError
        get_settings_func(
            "UPLOAD_FILTER_FUNC",
            "spkcspider.apps.spider.functions.allow_all_filter"
        )(ret["file"])
        return ret


class TextForm(forms.ModelForm):
    text = SanitizedHtmlField(localize=True)
    license_name = OpenChoiceField(
        label=_("License"), help_text=_("Select license"),
        choices=map(_extract_choice, LICENSE_CHOICES.items()),
        widget=LicenseChooserWidget(licenses=LICENSE_CHOICES)
    )
    sources = MultipleOpenChoiceField(
        required=False, initial=False,
        widget=ListWidget(
            item_label=_("Source")
        )
    )

    class Meta:
        model = TextFilet
        fields = [
            'text', 'push', 'editable_from',
            'license_name', 'license_url', 'sources'
        ]

        widgets = {
            "editable_from": Select2Widget(
                allow_multiple_selected=True,
                attrs={
                    "style": "min-width: 150px; width:100%"
                }
            ),
            "license_url": forms.URLInput(
                attrs={
                    "style": "width:100%"
                }
            )
        }

    class Media:
        js = [
            'spider_filets/licensechooser.js',
            'spider_filets/description_helper.js'
        ]

    def __init__(self, request, source, scope, initial=None, **kwargs):
        if initial is None:
            initial = {}
        if not getattr(kwargs.get("instance", None), "id", None):
            initial.setdefault(
                "license_name", DEFAULT_LICENSE_TEXT(source, request.user)
            )
        super().__init__(initial=initial, **kwargs)
        if scope in ("add", "update"):
            self.fields["editable_from"].help_text = \
                _(
                    "Allow editing from selected components. "
                    "Requires protection strength >=%s."
                ) % settings.MIN_STRENGTH_EVELATION

            query = models.Q(pk=self.instance.associated.usercomponent.pk)
            if scope == "update":
                query |= models.Q(
                    contents__references=self.instance.associated
                )
            query &= models.Q(strength__gte=settings.MIN_STRENGTH_EVELATION)
            query &= models.Q(strength__lt=9)
            self.fields["editable_from"].queryset = \
                self.fields["editable_from"].queryset.filter(query).distinct()
            return

        del self.fields["editable_from"]
        del self.fields["push"]
        self.fields["license_name"].editable = False
        self.fields["license_url"].editable = False

        allow_edit = scope == "update_guest"

        self.fields["text"].editable = allow_edit
        # sources stay enabled
        self.fields["sources"].editable = allow_edit


class RawTextForm(forms.ModelForm):
    name = forms.CharField()
    sources = MultipleOpenChoiceField(
        required=False, initial=False
    )

    class Meta:
        model = TextFilet
        fields = ['text', 'license_name', 'license_url', 'sources']

    def __init__(self, request, source=None, scope=None, **kwargs):
        super().__init__(**kwargs)
        self.fields['name'].initial = self.instance.associated.name
        setattr(self.fields['name'], "hashable", True)
        setattr(self.fields['text'], "hashable", True)
        # sources should not be hashed as they don't affect result
        setattr(self.fields['sources'], "hashable", False)
        setattr(self.fields['license_name'], "hashable", True)
        setattr(self.fields['license_url'], "hashable", True)



from django.contrib.auth.backends import ModelBackend
from django.http import Http404
from django.utils import timezone

from .models import UserComponent, Protection, TravelProtection, AuthToken
from .constants import ProtectionType, TravelLoginType


class SpiderTokenAuthBackend(ModelBackend):

    def authenticate(self, request, username=None,
                     protection_codes=None, **kwargs):
        """ Use protections for authentication"""
        tokenstring = request.GET.get("token", None)
        if not tokenstring:
            return
        travel = TravelProtection.objects.get_active().filter(
            associated_rel__usercomponent__user__username=username
        ).exclude(login_protection=TravelLoginType.none.value)

        is_fake = False

        if travel.exists():
            uc = UserComponent.objects.filter(
                user__username=username, name="fake_index"
            ).first(),
            is_fake = True
        else:
            uc = UserComponent.objects.filter(
                user__username=username, name="index"
            ).first()

        expire = timezone.now()-uc.token_duration
        # delete old token, so no confusion happen
        AuthToken.objects.filter(
            usercomponent=uc, created__lt=expire
        ).delete()
        if AuthToken.objects.filter(
            usercomponent=uc,
            token=tokenstring
        ).exists():
            request.session["is_fake"] = is_fake
            return uc.user


class SpiderAuthBackend(ModelBackend):

    def authenticate(self, request, username=None,
                     protection_codes=None, nospider=False, **kwargs):
        """ Use protections for authentication"""
        # disable SpiderAuthBackend backend (against recursion)
        if nospider:
            return
        travel = TravelProtection.objects.get_active().filter(
            associated_rel__usercomponent__user__username=username
        ).exclude(login_protection=TravelLoginType.none.value)
        uc = UserComponent.objects.filter(
            user__username=username, name="index"
        ).first()

        uc_fake = None
        if travel.exists():
            uc_fake = UserComponent.objects.filter(
                user__username=username, name="fake_index"
            ).first()
        try:
            if not uc:
                request.protections = Protection.authall(
                    request, scope="auth",
                    ptype=ProtectionType.authentication.value,
                    protection_codes=protection_codes
                )
                if type(request.protections) is int:  # should never happen
                    return None
            else:
                if uc_fake:
                    request.protections = uc_fake.auth(
                        request, scope="auth",
                        ptype=ProtectionType.authentication.value,
                        protection_codes=protection_codes
                    )
                    if type(request.protections) is int:
                        request.session["is_fake"] = True
                        return uc_fake.user
                # don't overwrite request.protections yet to serve fake version
                # in case the real login doesn't work either
                protections = uc.auth(
                    request, scope="auth",
                    ptype=ProtectionType.authentication.value,
                    protection_codes=protection_codes
                )
                if type(protections) is int:
                    request.protections = protections
                    request.session["is_fake"] = False
                    return uc.user
                # there was no fake so set protections
                if not uc_fake:
                    request.protections = protections
        except Http404:
            # for Http404 auth abort by protections (e.g. Random Fail)
            pass



"""spkcspider URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""

from django.conf.urls.i18n import i18n_patterns
from django.urls import path, include
from django.conf.urls.static import static
from django.contrib import admin
from django.views.generic.base import RedirectView
from django.conf import settings
from django.apps import apps
from spkcspider.apps.spider.views import ComponentPublicIndex
from spkcspider.apps.spider.helpers import get_settings_func
from spkcspider.apps.spider.functions import admin_login
from django.contrib.sitemaps import views as sitemaps_views
from django.views.decorators.cache import cache_page

from spkcspider.apps.spider.sitemaps import sitemaps

favicon_view = RedirectView.as_view(
    url='{}spider_base/favicon.svg'.format(settings.STATIC_URL), permanent=True
)
robots_view = RedirectView.as_view(
    url='{}spider_base/robots.txt'.format(settings.STATIC_URL), permanent=True
)

# disable admin login page
admin.site.login = lambda *args, **kwargs: admin_login(
    admin.site, *args, **kwargs
)
# default: allow only non faked user with superuser and staff permissions
admin.site.has_permission = lambda *args, **kwargs: get_settings_func(
    "HAS_ADMIN_PERMISSION_FUNC",
    "spkcspider.apps.spider.functions.has_admin_permission"
)(admin.site, *args, **kwargs)

urlpatterns = []
urlpatterns_i18n = [
    path('admin/', admin.site.urls),
    path(
        '',
        ComponentPublicIndex.as_view(
            is_home=True, template_name="spider_base/home.html"
        ),
        name="home"
    ),
]

for app in apps.get_app_configs():
    url_path = getattr(
        app, "spider_url_path",
        None
    )
    if not url_path:
        continue
    urlpatterns_i18n.append(
        path(
            url_path,
            include("{}.urls".format(app.name))
        )
    )


if getattr(settings, "SPIDER_LEGACY_REDIRECT", False):
    urlpatterns_i18n.insert(
        0,
        path(
            'spider/content/access/<path:token>/<slug:access>/',
            RedirectView.as_view(
                pattern_name='spider_base:ucontent-access',
                permanent=True
            )
        )
    )
    urlpatterns_i18n.insert(
        0,
        path(
            'spider/ucs/list/<path:token>/',
            RedirectView.as_view(
                pattern_name='spider_base:ucontent-list',
                permanent=True
            )
        )
    )


if getattr(settings, "USE_CAPTCHAS", False):
    urlpatterns.append(path(r'captcha/', include('captcha.urls')))

if 'django.contrib.flatpages' in settings.INSTALLED_APPS:
    urlpatterns_i18n.append(
        path('pages/', include('django.contrib.flatpages.urls'))
    )

urlpatterns += [
    # daily
    path('favicon.ico', cache_page(86400)(favicon_view)),
    path('robots.txt', cache_page(86400)(robots_view)),
    # daily
    path(
        'sitemap.xml',
        cache_page(86400)(
            sitemaps_views.index
        ),
        {
            'sitemaps': sitemaps,
            'sitemap_url_name': 'sitemaps'
        },
        name='django.contrib.sitemaps.views.index'
    ),
    # hourly
    path(
        'sitemap-<section>.xml',
        cache_page(3600)(
            sitemaps_views.sitemap
        ),
        {'sitemaps': sitemaps},
        name='sitemaps'
    ),
] + i18n_patterns(*urlpatterns_i18n, prefix_default_language=False)


if settings.DEBUG:
    urlpatterns += static(
        settings.MEDIA_URL, document_root=settings.MEDIA_ROOT
    )



# Generated by Django 2.1 on 2018-08-10 00:34

from django.db import migrations, models
import spkcspider.apps.spider_keys.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PublicKey',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('key', models.TextField(validators=[spkcspider.apps.spider_keys.models.valid_pkey_properties])),
                ('note', models.TextField(blank=True, default='', max_length=100)),
            ],
            options={
                'abstract': False,
                'default_permissions': [],
            },
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-09 15:05

from django.db import migrations, models
import spkcspider.apps.spider_accounts.models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_accounts', '0003_spideruser_quota'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='spideruser',
            name='quota',
        ),
        migrations.AddField(
            model_name='spideruser',
            name='quota_local',
            field=models.PositiveIntegerField(blank=True, default=spkcspider.apps.spider_accounts.models.default_quota_spider_user_local, help_text='Quota in Bytes, null for no limit', null=True),
        ),
        migrations.AddField(
            model_name='spideruser',
            name='quota_remote',
            field=models.PositiveIntegerField(blank=True, default=spkcspider.apps.spider_accounts.models.default_quota_spider_user_remote, help_text='Quota in Bytes, null for no limit', null=True),
        ),
    ]



__all__ = ["literalize", "uriref", "spkc_namespace"]

from django import template
from django.forms import BoundField, Field

from rdflib import Literal
from rdflib.namespace import XSD, RDF, URIRef

from ..constants import spkcgraph
from ..helpers import merge_get_url

register = template.Library()


@register.filter()
def uriref(path):
    return URIRef(path)


@register.filter()
def is_uriref(value):
    return isinstance(value, URIRef)


@register.simple_tag(takes_context=True)
def action_view(context):
    token = getattr(context["request"], "auth_token", None)
    if token:
        token = token.token
    url2 = merge_get_url(
        context["hostpart"] + context["request"].path, token=token
    )
    return Literal(url2, datatype=XSD.anyURI)


@register.simple_tag()
def literalize(ob, datatype=None, use_uriref=None):
    if isinstance(ob, BoundField):
        if not datatype:
            datatype = getattr(ob.field, "spkc_datatype", None)
        if use_uriref is None:
            use_uriref = getattr(ob.field, "spkc_use_uriref", None)
        ob = ob.value()
    elif isinstance(datatype, BoundField):
        if use_uriref is None:
            use_uriref = getattr(datatype.field, "spkc_use_uriref", None)
        datatype = getattr(datatype.field, "spkc_datatype", None)
    elif isinstance(datatype, Field):
        if use_uriref is None:
            use_uriref = getattr(datatype, "spkc_use_uriref", None)
        datatype = getattr(datatype, "spkc_datatype", None)
    if ob is None:
        return RDF.nil
    if hasattr(ob, "get_absolute_url"):
        if not datatype:
            datatype = spkcgraph["hashableURI"]
        if use_uriref is None:
            use_uriref = True
        ob = ob.get_absolute_url()
    elif isinstance(ob, str) and not datatype:
        datatype = XSD.string
    if use_uriref:
        return URIRef(ob)
    return Literal(ob, datatype=datatype)


@register.simple_tag()
def hashable_literalize(field):
    if isinstance(field, BoundField):
        return literalize(getattr(field.field, "hashable", False))
    else:
        return literalize(getattr(field, "hashable", False))


@register.simple_tag()
def spkc_namespace(sub=None):
    if sub:
        return spkcgraph[sub]
    return spkcgraph



# Generated by Django 2.1.5 on 2019-01-27 14:47

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import jsonfield.fields
import spkcspider.apps.spider.models.contents
import spkcspider.apps.spider.models.protections


class Migration(migrations.Migration):

    replaces = [('spider_base', '0002_20181231_squashed'), ('spider_base', '0003_remove_usercomponent_avail_features'), ('spider_base', '0004_auto_20190103_0156'), ('spider_base', '0005_auto_20190109_1505'), ('spider_base', '0006_assignedcontent_priority'), ('spider_base', '0007_linkcontent_push'), ('spider_base', '0008_authtoken_referrer'), ('spider_base', '0009_authtoken_persist'), ('spider_base', '0010_usercomponent_can_auth'), ('spider_base', '0011_auto_20190117_1256'), ('spider_base', '0012_auto_20190118_1017'), ('spider_base', '0013_assignedcontent_persist_token')]

    dependencies = [
        ('spider_base', '0001_initial'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='strength',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MaxValueValidator(10)]),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='strength_link',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MaxValueValidator(11)]),
        ),
        migrations.AddField(
            model_name='contentvariant',
            name='strength',
            field=models.PositiveSmallIntegerField(default=0, validators=[django.core.validators.MaxValueValidator(10)]),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='strength',
            field=models.PositiveSmallIntegerField(default=0, editable=False, validators=[django.core.validators.MaxValueValidator(10)]),
        ),
        migrations.AlterField(
            model_name='contentvariant',
            name='ctype',
            field=models.CharField(default='', max_length=10),
        ),
        migrations.AlterField(
            model_name='contentvariant',
            name='name',
            field=models.SlugField(unique=True),
        ),
        migrations.CreateModel(
            name='TravelProtection',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('active', models.BooleanField(default=False)),
                ('start', models.DateTimeField(default=spkcspider.apps.spider.models.contents.default_start)),
                ('stop', models.DateTimeField(default=spkcspider.apps.spider.models.contents.default_stop, null=True)),
                ('self_protection', models.BooleanField(default=True, help_text='\n    Disallows user to disable travel protection if active.\n    Can be used in connection with "secret" to allow unlocking via secret\n')),
                ('login_protection', models.CharField(choices=[('a', 'No Login protection')], default='a', help_text='\n    No Login Protection: normal, default\n    Fake Login: fake login and index (experimental)\n    Wipe: Wipe protected content,\n    except they are protected by a deletion period\n    Wipe User: destroy user on login\n\n\n    <div>\n        Danger: every option other than: "No Login Protection" can screw you.\n        "Fake Login" can trap you in a parallel reality\n    </div>\n', max_length=10)),
                ('secret', models.SlugField(default='', max_length=120)),
            ],
            options={
                'abstract': False,
                'default_permissions': [],
            },
        ),
        migrations.AlterField(
            model_name='assignedprotection',
            name='protection',
            field=models.ForeignKey(limit_choices_to=spkcspider.apps.spider.models.protections.get_limit_choices_assigned_protection, on_delete=django.db.models.deletion.CASCADE, related_name='assigned', to='spider_base.Protection'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='deletion_requested',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='travelprotection',
            name='disallow',
            field=models.ManyToManyField(blank=True, related_name='travel_protected', to='spider_base.UserComponent'),
        ),
        migrations.AlterModelOptions(
            name='usercomponent',
            options={'permissions': [('can_feature', 'Can feature User Components')]},
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='featured',
            field=models.BooleanField(default=False, help_text='Appears as featured on "home" page'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='name',
            field=models.SlugField(allow_unicode=True, help_text='\nName of the component.<br/>\nNote: there are special named components\nwith different protection types and scopes.<br/>\nMost prominent: "index" for authentication\n'),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='description',
            field=models.TextField(blank=True, default='', help_text='Description of user component.'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='fake_id',
            field=models.BigIntegerField(editable=False, null=True),
        ),
        migrations.RemoveField(
            model_name='travelprotection',
            name='secret',
        ),
        migrations.RemoveField(
            model_name='travelprotection',
            name='self_protection',
        ),
        migrations.AddField(
            model_name='travelprotection',
            name='hashed_secret',
            field=models.CharField(max_length=128, null=True),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='references',
            field=models.ManyToManyField(editable=False, related_name='referenced_by', to='spider_base.AssignedContent'),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='created_by_special_user',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='extra',
            field=jsonfield.fields.JSONField(blank=True, default=dict),
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='login_protection',
            field=models.CharField(choices=[('a', 'No Login protection'), ('b', 'Fake login'), ('c', 'Wipe'), ('d', 'Wipe User')], default='a', help_text='\n    No Login Protection: normal, default\n    Fake Login: fake login and index (experimental)\n    Wipe: Wipe protected content,\n    except they are protected by a deletion period\n    Wipe User: destroy user on login\n\n\n    <div>\n        Danger: every option other than: "No Login Protection" can screw you.\n        "Fake Login" can trap you in a parallel reality\n    </div>\n', max_length=10),
        ),
        migrations.AddField(
            model_name='travelprotection',
            name='is_fake',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='travelprotection',
            name='login_protection',
            field=models.CharField(choices=[('a', 'No Login protection'), ('b', 'Fake login'), ('c', 'Wipe'), ('d', 'Wipe User')], default='a', help_text='No Login Protection: normal, default<br/>Fake Login: fake login and index<br/>Wipe: Wipe protected content, except they are protected by a deletion period<br/>Wipe User: destroy user on login', max_length=10),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='name',
            field=models.SlugField(allow_unicode=True, help_text='Name of the component.<br/>Note: there are special named components with different protection types and scopes.<br/>Most prominent: "index" for authentication'),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='features',
            field=models.ManyToManyField(blank=True, limit_choices_to=models.Q(ctype__contains='a'), related_name='supports', to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='ctype',
            field=models.ForeignKey(editable=False, limit_choices_to=models.Q(_negated=True, ctype__contains='a'), null=True, on_delete=django.db.models.deletion.SET_NULL, to='spider_base.ContentVariant'),
        ),
        migrations.AlterField(
            model_name='authtoken',
            name='token',
            field=models.SlugField(max_length=136, unique=True),
        ),
        migrations.AlterUniqueTogether(
            name='authtoken',
            unique_together=set(),
        ),
        migrations.AlterField(
            model_name='assignedcontent',
            name='ctype',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to='spider_base.ContentVariant'),
        ),
        migrations.RenameField(
            model_name='userinfo',
            old_name='used_space',
            new_name='used_space_local',
        ),
        migrations.AddField(
            model_name='userinfo',
            name='used_space_remote',
            field=models.BigIntegerField(default=0, editable=False),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='priority',
            field=models.SmallIntegerField(blank=True, default=0),
        ),
        migrations.AddField(
            model_name='linkcontent',
            name='push',
            field=models.BooleanField(blank=True, default=False, help_text='Improve ranking of this Link.'),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='referrer',
            field=models.URLField(blank=True, max_length=400, null=True),
        ),
        migrations.AddField(
            model_name='authtoken',
            name='persist',
            field=models.BigIntegerField(blank=True, db_index=True, default=-1),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='can_auth',
            field=models.BooleanField(default=False, editable=False),
        ),
        migrations.AddField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, help_text='Select main identyifing anchor. Also used for attaching persisting tokens (elsewise they are attached to component)', limit_choices_to={'info__contains': '\nanchor\n'}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='primary_anchor_for', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='linkcontent',
            name='content',
            field=models.ForeignKey(limit_choices_to={'strength_link__lte': 10}, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='public',
            field=models.BooleanField(default=False, help_text='Is public? Is listed and searchable?<br/>Note: This field is maybe not deactivatable because of assigned content'),
        ),
        migrations.AddField(
            model_name='assignedcontent',
            name='persist_token',
            field=models.ForeignKey(blank=True, limit_choices_to={'persist__gte': 0}, null=True, on_delete=django.db.models.deletion.CASCADE, to='spider_base.AuthToken'),
        ),
    ]




def UpdateDefaultLayouts(sender, **kwargs):
    # provided apps argument lacks model function support
    # so use this
    from django.apps import apps
    from .layouts import initialize_layouts
    initialize_layouts(apps)



__all__ = ["FlatpageItemForm"]

import copy
from django.contrib.flatpages.forms import FlatpageForm
from django.utils.translation import gettext_lazy as _
from ..widgets import TrumbowygWidget

_help_text = _(
    """Example: '/about/contact/'. Make sure to have leading and trailing slashes.
    <br/>Special Urls (ordered by url):<br/>
    /home/heading/*/ : flatpages used for heading on frontpage <br/>
    /home/main/*/ : flatpages used for general information on frontpage <br/>
    /gfooter/*/ : flatpage links with general informations (links will be rendered on every page)
    """  # noqa: E501
)


class FlatpageItemForm(FlatpageForm):
    url = copy.copy(FlatpageForm.base_fields["url"])
    url.help_text = _help_text

    class Meta(FlatpageForm.Meta):
        widgets = {
            "content": TrumbowygWidget()
        }



__all__ = (
    "UserTestMixin", "UCTestMixin", "EntityDeletionMixin", "ReferrerMixin"
)

import hashlib
import logging
from urllib.parse import quote_plus

from datetime import timedelta

from django.contrib.auth.mixins import AccessMixin
from django.shortcuts import get_object_or_404
from django.contrib.auth import get_user_model, REDIRECT_FIELD_NAME
from django.http import (
    HttpResponseRedirect, HttpResponseServerError, HttpResponse
)
from django.http.response import HttpResponseBase
from django.db import transaction
from django.db.models import Q
from django.utils import timezone
from django.conf import settings
from django.urls import reverse_lazy
from django.utils.translation import gettext

try:
    from ratelimit.core import is_ratelimited, get_usage, user_or_ip
except ImportError:
    from ratelimit.utils import (
        is_ratelimited, user_or_ip, get_usage_count as get_usage
    )

import requests
import certifi

from ..helpers import merge_get_url, get_settings_func, get_hashob
from ..constants import VariantType, TokenCreationError, ProtectionType
from ..conf import VALID_INTENTIONS, VALID_SUB_INTENTIONS
from ..models import UserComponent, AuthToken, ReferrerObject


class UserTestMixin(AccessMixin):
    preserved_GET_parameters = set(["token", "protection"])
    login_url = reverse_lazy(getattr(
        settings,
        "LOGIN_URL",
        "auth:login"
    ))

    def dispatch_extra(self, request, *args, **kwargs):
        return None

    def dispatch(self, request, *args, **kwargs):
        _ = gettext
        self.request.is_owner = getattr(self.request, "is_owner", False)
        self.request.is_special_user = \
            getattr(self.request, "is_special_user", False)
        self.request.is_staff = getattr(self.request, "is_staff", False)
        self.request.auth_token = getattr(self.request, "auth_token", None)
        try:
            user_test_result = self.test_func()
        except TokenCreationError:
            logging.exception("Token creation failed")
            return HttpResponseServerError(
                _("Token creation failed, try again")
            )
        if isinstance(user_test_result, HttpResponseBase):
            return user_test_result
        elif not user_test_result:
            return self.handle_no_permission()

        ret = self.dispatch_extra(request, *args, **kwargs)
        if ret:
            return ret
        return super().dispatch(request, *args, **kwargs)

    def sanitize_GET(self):
        GET = self.request.GET.copy()
        for key in list(GET.keys()):
            if key not in self.preserved_GET_parameters:
                GET.pop(key, None)
        return GET

    def get_context_data(self, **kwargs):
        kwargs["raw_update_type"] = VariantType.raw_update.value
        kwargs["feature_type"] = VariantType.component_feature.value
        kwargs["hostpart"] = "{}://{}".format(
            self.request.scheme, self.request.get_host()
        )
        kwargs["spider_GET"] = self.sanitize_GET()
        return super().get_context_data(**kwargs)

    # by default only owner with login can access view
    def test_func(self):
        if self.has_special_access(
            user_by_login=True
        ):
            return True
        return False

    def replace_token(self):
        GET = self.request.GET.copy()
        GET["token"] = self.request.auth_token.token
        return HttpResponseRedirect(
            redirect_to="?".join((self.request.path, GET.urlencode()))
        )

    def create_token(self, special_user=None, extra=None):
        d = {
            "usercomponent": self.usercomponent,
            "session_key": None,
            "created_by_special_user": special_user,
            "extra": {}
        }
        if "token" not in self.request.GET:
            d["session_key"] = self.request.session.session_key

        token = AuthToken(**d)
        if extra:
            token.extra.update(extra)
        token.save()
        return token

    def create_admin_token(self):
        expire = timezone.now()-self.usercomponent.token_duration
        # delete old token, so no confusion happen
        self.usercomponent.authtokens.filter(
            created__lt=expire
        ).delete()
        # delete tokens from old sessions
        self.usercomponent.authtokens.exclude(
            session_key=self.request.session.session_key,
        ).filter(created_by_special_user=self.request.user).delete()

        # use session_key, causes deletion on logout
        token = self.usercomponent.authtokens.filter(
            session_key=self.request.session.session_key
        ).first()
        if token:
            return token
        return self.create_token(
            self.request.user,
            extra={
                "strength": 10
            }
        )

    def remove_old_tokens(self, expire=None):
        if not expire:
            expire = timezone.now()-self.usercomponent.token_duration
        return self.usercomponent.authtokens.filter(
            created__lt=expire, persist=-1
        ).delete()

    def test_token(self, minstrength=0, force_token=False):
        expire = timezone.now()-self.usercomponent.token_duration
        no_token = not force_token and self.usercomponent.required_passes == 0
        ptype = ProtectionType.access_control.value
        if minstrength >= 4:
            no_token = False
            ptype = ProtectionType.authentication.value

        # delete old token, so no confusion happen
        self.remove_old_tokens(expire)

        # only valid tokens here
        tokenstring = self.request.GET.get("token", None)
        token = None
        if tokenstring:
            # find by tokenstring
            token = self.usercomponent.authtokens.filter(
                token=tokenstring
            ).first()
        elif self.request.session.session_key:
            # use session_key
            token = self.usercomponent.authtokens.filter(
                session_key=self.request.session.session_key
            ).first()
        elif not no_token:
            # generate session key if it not exist and token is required
            self.request.session.cycle_key()
        if token and token.extra.get("prot_strength", 0) >= minstrength:
            self.request.token_expires = \
                token.created+self.usercomponent.token_duration
            # case will never enter
            # if not token.session_key and "token" not in self.request.GET:
            #     return self.replace_token()
            if token.extra.get("prot_strength", 0) >= 4:
                self.request.is_special_user = True
                self.request.is_owner = True
            self.request.auth_token = token
            return True

        # if result is impossible and token invalid try to login
        if minstrength >= 4 and not self.usercomponent.can_auth:
            # remove token and redirect
            # login_url can be also on a different host => merge_get_url
            target = "{}?{}={}".format(
                self.get_login_url(),
                REDIRECT_FIELD_NAME,
                quote_plus(
                    merge_get_url(
                        self.request.get_full_path(),
                        token=None
                    )
                )
            )
            return HttpResponseRedirect(redirect_to=target)

        protection_codes = None
        if "protection" in self.request.GET:
            protection_codes = self.request.GET.getlist("protection")
        # execute protections for side effects even no_token
        self.request.protections = self.usercomponent.auth(
            request=self.request, scope=self.scope,
            protection_codes=protection_codes, ptype=ptype
        )
        if (
            type(self.request.protections) is int and  # because: False==0
            self.request.protections >= minstrength
        ):
            # generate no token if not required
            if no_token:
                return True

            token = self.create_token(
                extra={
                    "strength": self.usercomponent.strength,
                    "prot_strength": self.request.protections
                }
            )

            if token.extra["prot_strength"] >= 4:
                self.request.is_special_user = True
                self.request.is_owner = True

            self.request.token_expires = \
                token.created+self.usercomponent.token_duration
            self.request.auth_token = token
            if "token" in self.request.GET:
                return self.replace_token()
            return True
        return False

    def has_special_access(
        self, user_by_login=True, user_by_token=False,
        staff=False, superuser=False
    ):
        if not hasattr(self, "usercomponent"):
            self.usercomponent = self.get_usercomponent()
        if user_by_login and self.request.user == self.usercomponent.user:
            self.request.is_owner = True
            self.request.is_special_user = True
            return True
        if user_by_token and self.test_token(4) is True:
            return True

        # remove user special state if is_fake
        if self.request.session.get("is_fake", False):
            return False
        if superuser and self.request.user.is_superuser:
            self.request.is_special_user = True
            self.request.is_staff = True
            return True
        if staff and self.request.user.is_staff:
            if type(staff) is bool:
                self.request.is_special_user = True
                self.request.is_staff = True
                return True
            if not isinstance(staff, (tuple, list, set)):
                staff = (staff,)

            if not all(
                map(self.request.user.has_perm, staff)
            ):
                return False
            self.request.is_special_user = True
            self.request.is_staff = True
            return True
        return False

    def get_user(self):
        """ Get user from user field or request """
        if (
                "user" not in self.kwargs and
                self.request.user.is_authenticated
           ):
            return self.request.user

        model = get_user_model()
        margs = {model.USERNAME_FIELD: None}
        margs[model.USERNAME_FIELD] = self.kwargs.get("user", None)
        return get_object_or_404(
            model.objects.select_related("spider_info"),
            **margs
        )

    def get_usercomponent(self):
        query = {
            "token": self.kwargs["token"]
        }
        return get_object_or_404(
            UserComponent.objects.prefetch_related(
                "authtokens", "protections"
            ),
            **query
        )

    def handle_no_permission(self):
        # in case no protections are used (e.g. add content)
        p = getattr(self.request, "protections", False)
        if not bool(p):
            # return 403
            return super().handle_no_permission()
        # should be never true here
        assert(p is not True)
        context = {
            "spider_GET": self.sanitize_GET(),
            "LOGIN_URL": self.get_login_url(),
            "scope": getattr(self, "scope", None),
            "uc": self.usercomponent,
            "object": getattr(self, "object", None),
            "is_public_view": self.usercomponent.public
        }
        return self.response_class(
            request=self.request,
            template=self.get_noperm_template_names(),
            # render with own context; get_context_data may breaks stuff or
            # disclose informations
            context=context,
            using=self.template_engine,
            content_type=self.content_type
        )

    def get_noperm_template_names(self):
        return "spider_base/protections/protections.html"


class UCTestMixin(UserTestMixin):
    usercomponent = None

    def dispatch(self, request, *args, **kwargs):
        self.usercomponent = self.get_usercomponent()
        return super(UCTestMixin, self).dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        # for protections & contents
        return self.get(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        # for protections & contents
        return self.get(request, *args, **kwargs)


class ReferrerMixin(object):
    allow_domain_mode = False

    def get_context_data(self, **kwargs):
        kwargs["token_strength"] = None
        # will be overwritten in referring path so there is no interference
        kwargs["referrer"] = None
        kwargs["intentions"] = set()
        if self.request.auth_token:
            kwargs["referrer"] = self.request.auth_token.referrer
            kwargs["token_strength"] = self.request.auth_token.extra.get(
                "strength", None
            )
            kwargs["intentions"].update(self.request.auth_token.extra.get(
                "intentions", []
            ))
        return super().get_context_data(**kwargs)

    def test_token(self, minstrength, force_token=False):
        if "intention" in self.request.GET or "referrer" in self.request.GET:
            # validate early, before auth
            intentions = set(self.request.GET.getlist("intention"))
            if not VALID_INTENTIONS.issuperset(
                intentions
            ):
                return HttpResponse(
                    "invalid intentions", status=400
                )
            if "domain" in intentions:

                if not self.clean_domain_upgrade(
                    {"intentions": intentions},
                    False
                ):
                    return HttpResponse(
                        "invalid domain upgrade", status=400
                    )
                # requires token
                force_token = True
            else:
                # maximal one main intention
                if len(intentions.difference(VALID_SUB_INTENTIONS)) > 1:
                    return HttpResponse(
                        "invalid intentions", status=400
                    )
                minstrength = 4
        return super().test_token(minstrength, force_token)

    def refer_with_post(self, context, token):
        # application/x-www-form-urlencoded is best here,
        # for beeing compatible to most webservers
        # client side rdf is no problem
        # NOTE: csrf must be disabled or use csrf token from GET,
        #       here is no way to know the token value

        h = hashlib.sha256(context["referrer"].encode("utf8")).hexdigest()

        def h_fun(*a):
            return h
        # rate limit on errors
        if is_ratelimited(
            request=self.request,
            fn=self.refer_with_post,
            key=h_fun,
            rate=settings.SPIDER_DOMAIN_ERROR_RATE,
            increment=False
        ):
            return HttpResponseRedirect(
                redirect_to=merge_get_url(
                    context["referrer"],
                    status="post_failed",
                    error="error_rate_limit"
                )
            )
        try:
            d = {
                "token": token.token,
                "hash_algorithm": settings.SPIDER_HASH_ALGORITHM.name,
                "renew": "false"
            }
            if context["payload"]:
                d["payload"] = context["payload"]

            ret = requests.post(
                context["referrer"],
                data=d,
                timeout=settings.SPIDER_REQUESTS_TIMEOUT,
                headers={
                    "Referer": merge_get_url(
                        "%s%s" % (
                            context["hostpart"],
                            self.request.path
                        )
                        # sending full url not required anymore, payload
                        # token=None, referrer=None, raw=None, intention=None,
                        # sl=None, payload=None
                    )
                },
                verify=certifi.where()
            )
            ret.raise_for_status()
        except requests.exceptions.SSLError as exc:
            logging.info(
                "referrer: \"%s\" has a broken ssl configuration",
                context["referrer"], exc_info=exc
            )
            return HttpResponseRedirect(
                redirect_to=merge_get_url(
                    context["referrer"],
                    status="post_failed",
                    error="ssl"
                )
            )
        except Exception as exc:
            apply_error_limit = False
            if isinstance(
                exc, (
                    requests.exceptions.ConnectionError,
                    requests.exceptions.Timeout
                )
            ):
                apply_error_limit = True
            elif (
                isinstance(exc, requests.exceptions.HTTPError) and
                exc.response.status_code >= 500
            ):
                apply_error_limit = True
            if apply_error_limit:
                get_usage(
                    request=self.request,
                    fn=self.refer_with_post,
                    key=h_fun,
                    rate=settings.SPIDER_DOMAIN_ERROR_RATE,
                    increment=True
                )
            logging.info(
                "post failed: \"%s\" failed",
                context["referrer"], exc_info=exc
            )
            return HttpResponseRedirect(
                redirect_to=merge_get_url(
                    context["referrer"],
                    status="post_failed",
                    error="other"
                )
            )
        context["post_success"] = True
        h = get_hashob()
        h.update(token.token.encode("ascii", "ignore"))
        return HttpResponseRedirect(
            redirect_to=merge_get_url(
                context["referrer"],
                status="success",
                hash=h.finalize().hex()
            )
        )

    def refer_with_get(self, context, token):
        return HttpResponseRedirect(
            redirect_to=merge_get_url(
                context["referrer"],
                token=token.token,
                payload=context["payload"]
            )
        )

    def _get_clean_domain_upgrade_key(self, group, request):
        return "{}:{}".format(
            self.usercomponent.id,
            user_or_ip(request)
        )

    def clean_domain_upgrade(self, context, token):
        if "referrer" not in self.request.GET:
            return False
        # domain mode must be used alone
        if len(context["intentions"]) > 1:
            return False
        if not context["intentions"].issubset(VALID_INTENTIONS):
            return False

        if not getattr(self.request, "_clean_domain_upgrade_checked", False):
            if is_ratelimited(
                request=self.request,
                fn=self.clean_domain_upgrade,
                key=self._get_clean_domain_upgrade_key,
                rate=settings.SPIDER_DOMAIN_UPDATE_RATE,
                increment=True
            ):
                return False

            setattr(self.request, "_clean_domain_upgrade_checked", True)

        # False for really no token
        if token is False:
            return True
        if not token or token.created_by_special_user:
            return False
        return True

    def clean_refer_intentions(self, context, token=None):
        # Only owner can use other intentions than domain
        if not self.request.is_owner:
            return False
        # Second error: invalid intentions
        #  this is the second time the validation will be executed
        #    in case test_token path is used
        #  this is the first time the validation will be executed
        #    in case has_special_access path is used
        if not context["intentions"].issubset(VALID_INTENTIONS):
            return False

        # auth is only for self.requesting component auth
        if "auth" in context["intentions"]:
            return False
        # maximal one main intention
        if len(context["intentions"].difference(VALID_SUB_INTENTIONS)) > 1:
            return False
        # "persist" or default can be serverless other intentions not
        #  this way rogue client based attacks are prevented
        if "persist" in context["intentions"]:
            if not self.usercomponent.features.filter(
                name="Persistence"
            ).exists():
                return False
        else:
            if context["is_serverless"] and len(context["intentions"]) != 1:
                return False

        if not token:
            return True

        ####### with token ########  # noqa: 266E
        if "persist" in context["intentions"]:
            # cannot add sl intention
            if "intentions" in token.extra:
                if "sl" in context["intentions"].difference(
                    token.extra["intentions"]
                ):
                    return False
            # set persist = true, (false=-1)
            token.persist = 0
            # if possible, pin to anchor
            if self.usercomponent.primary_anchor:
                token.persist = self.usercomponent.primary_anchor.id
        else:
            # check if token was reused if not persisted
            if token.referrer:
                return False

        if "initial_referrer_url" not in token.extra:
            token.extra["initial_referrer_url"] = "{}://{}{}".format(
                self.request.scheme,
                self.request.get_host(),
                self.request.path
            )
        return True

    def handle_referrer(self):
        _ = gettext
        if (
            self.request.user != self.usercomponent.user and
            not self.request.auth_token
        ):
            return HttpResponseRedirect(
                redirect_to="{}?{}={}".format(
                    self.get_login_url(),
                    REDIRECT_FIELD_NAME,
                    quote_plus(
                        merge_get_url(
                            self.request.get_full_path(),
                            token=None
                        )
                    )
                )
            )

        context = self.get_context_data()
        context["intentions"] = set(self.request.GET.getlist("intention"))
        context["payload"] = self.request.GET.get("payload", None)
        context["is_serverless"] = "sl" in context["intentions"]
        context["referrer"] = merge_get_url(self.request.GET["referrer"])
        context["old_search"] = []
        context["object_list"] = self.object_list
        if not get_settings_func(
            "SPIDER_URL_VALIDATOR",
            "spkcspider.apps.spider.functions.validate_url_default"
        )(context["referrer"], self):
            return HttpResponse(
                status=400,
                content=_('Insecure url: %(url)s') % {
                    "url": context["referrer"]
                }
            )
        delete_auth_token = False

        action = self.request.POST.get("action", None)
        if "domain" in context["intentions"]:
            # domain mode only possible for token without user
            token = self.request.auth_token
            if not self.allow_domain_mode:
                return HttpResponse(
                    status=400,
                    content='domain mode disallowed'
                )
            if not self.clean_domain_upgrade(context, token):
                return HttpResponse(
                    status=400,
                    content='Invalid token'
                )
            token.create_auth_token()
            token.referrer = ReferrerObject.objects.get_or_create(
                url=context["referrer"]
            )[0]
            token.extra["intentions"] = list(context["intentions"])
            try:
                token.save()
            except TokenCreationError:
                logging.exception("Token creation failed")
                return HttpResponseServerError(
                    "Token creation failed, try again"
                )
            context["post_success"] = False
            ret = self.refer_with_post(context, token)
            if not context["post_success"]:
                token.delete()
            return ret
        elif action == "confirm":
            token = None
            hasoldtoken = False
            # if persist try to find old token
            if "persist" in context["intentions"]:
                persistfind = Q(persist=0, usercomponent=self.usercomponent)
                persistfind |= Q(
                    persist__in=self.usercomponent.contents.filter(
                        info__contains="\x1eanchor\x1e"
                    ).values_list("id", flat=True)
                )
                token = AuthToken.objects.filter(
                    persistfind,
                    referrer__url=context["referrer"]
                ).first()
                if token:
                    hasoldtoken = True
                    # migrate usercomponent
                    token.usercomponent = self.usercomponent

            # create only new token when admin token and not persisted token
            if self.usercomponent.user == self.request.user:
                # boost strength if user is owner
                # token is not auth_token
                if token:
                    token.extra["strength"] = 10
                    # self.request new token
                    token.create_auth_token()
                else:
                    token = AuthToken(
                        usercomponent=self.usercomponent,
                        extra={
                            "strength": 10
                        }
                    )
            else:
                # either reuse persistent token with auth token tokenstring
                # or just reuse auth token
                if token:
                    # slate auth token for destruction
                    delete_auth_token = True
                    # steal token value
                    token.token = self.request.auth_token.token
                else:
                    # repurpose token
                    # NOTE: one token, one referrer
                    token = self.request.auth_token

            # set to zero as prot_strength can elevate perms
            token.extra["prot_strength"] = 0
            token.extra["intentions"] = list(context["intentions"])

            if not self.clean_refer_intentions(context, token):
                return HttpResponseRedirect(
                    redirect_to=merge_get_url(
                        context["referrer"],
                        error="intentions_incorrect"
                    )
                )

            token.extra["filter"] = self.request.POST.getlist("search")
            if "live" in context["intentions"]:
                token.extra.pop("ids", None)
            else:
                token.extra["ids"] = list(
                    self.object_list.values_list("id", flat=True)
                )
            token.referrer = ReferrerObject.objects.get_or_create(
                url=context["referrer"]
            )[0]
            # after cleanup, save
            try:
                with transaction.atomic():
                    # must be done here, elsewise other token can (unlikely)
                    # take token as it is free for a short time, better be safe
                    if delete_auth_token:
                        # delete old token
                        self.request.auth_token.delete()
                    token.save()
            except TokenCreationError:
                logging.exception("Token creation failed")
                return HttpResponseServerError(
                    _("Token creation failed, try again")
                )

            if context["is_serverless"]:
                context["post_success"] = True
                ret = self.refer_with_get(context, token)
            else:
                context["post_success"] = False
                ret = self.refer_with_post(context, token)
            if not context["post_success"] and not hasoldtoken:
                token.delete()
            return ret

        elif action == "cancel":
            return HttpResponseRedirect(
                redirect_to=merge_get_url(
                    context["referrer"],
                    status="canceled",
                    payload=context["payload"]
                )
            )
        else:
            token = None
            oldtoken = None
            # use later reused token early
            if self.usercomponent.user != self.request.user:
                token = self.request.auth_token
            # don't re-add search parameters, only initialize
            if (
                self.request.method != "POST" and
                "persist" in context["intentions"]
            ):
                persistfind = Q(persist=0, usercomponent=self.usercomponent)
                persistfind |= Q(
                    persist__in=self.usercomponent.contents.filter(
                        info__contains="\x1eanchor\x1e"
                    ).values_list("id", flat=True)
                )
                oldtoken = AuthToken.objects.filter(
                    persistfind,
                    referrer__url=context["referrer"],
                ).first()

            if oldtoken:
                context["old_search"] = oldtoken.extra.get("search", [])
            if not self.clean_refer_intentions(context, token):
                return HttpResponse(
                    status=400,
                    content=_('Error: intentions incorrect')
                )
            return self.response_class(
                request=self.request,
                template=self.get_referrer_template_names(),
                context=context,
                using=self.template_engine,
                content_type=self.content_type
            )

    def get_referrer_template_names(self):
        return "spider_base/protections/referring.html"


class EntityDeletionMixin(UserTestMixin):
    object = None
    http_method_names = ['get', 'post', 'delete']

    def get_context_data(self, **kwargs):
        _time = self.get_required_timedelta()
        if _time and self.object.deletion_requested:
            now = timezone.now()
            if self.object.deletion_requested + _time >= now:
                kwargs["remaining"] = timedelta(seconds=0)
            else:
                kwargs["remaining"] = self.object.deletion_requested+_time-now
        return super().get_context_data(**kwargs)

    def get_required_timedelta(self):
        _time = self.object.deletion_period
        if _time:
            _time = timedelta(seconds=_time)
        else:
            _time = timedelta(seconds=0)
        return _time

    def options(self, request, *args, **kwargs):
        ret = super().options()
        ret["Access-Control-Allow-Origin"] = self.request.get_host()
        ret["Access-Control-Allow-Methods"] = "POST, GET, DELETE, OPTIONS"
        return ret

    def delete(self, request, *args, **kwargs):
        _time = self.get_required_timedelta()
        if _time:
            now = timezone.now()
            if self.object.deletion_requested:
                if self.object.deletion_requested+_time >= now:
                    return self.get(request, *args, **kwargs)
            else:
                self.object.deletion_requested = now
                self.object.save()
                return self.get(request, *args, **kwargs)
        self.object.delete()
        return HttpResponseRedirect(self.get_success_url())

    def post(self, request, *args, **kwargs):
        # because forms are screwed (delete not possible)
        # UPDATE: delete works if allowed in CORS
        if request.POST.get("action") == "reset":
            return self.reset(request, *args, **kwargs)
        elif request.POST.get("action") == "delete":
            return self.delete(request, *args, **kwargs)
        return super().get(request, *args, **kwargs)

    def reset(self, request, *args, **kwargs):
        self.object.deletion_requested = None
        self.object.save(update_fields=["deletion_requested"])
        return HttpResponseRedirect(self.get_success_url())



# Generated by Django 2.1.5 on 2019-01-15 01:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0008_authtoken_referrer'),
    ]

    operations = [
        migrations.AddField(
            model_name='authtoken',
            name='persist',
            field=models.BooleanField(blank=True, db_index=True, default=False),
        ),
    ]



from django.urls import path
from django.contrib.auth.decorators import login_required


from .views import (
    ComponentIndex, ComponentPublicIndex, ComponentCreate,
    ComponentUpdate, ComponentDelete
)

from .views import (
    ContentAdd, ContentIndex, ContentAccess, ContentDelete
)

from .views import (
    TokenDelete, TokenDeletionRequest, TokenRenewal
)

app_name = "spider_base"

# token: path: token can simulate path structures (legacy)

# components plural: most components url retrieve multiple items
#    one "component"-url for single retrievals is confusing

urlpatterns = [
    path(
        'components/export/',
        login_required(ComponentIndex.as_view(scope="export")),
        name='ucomponent-export'
    ),
    path(
        'components/slug:user>/export/',
        login_required(ComponentIndex.as_view(scope="export")),
        name='ucomponent-export'
    ),
    path(
        'components/list/',
        login_required(ComponentIndex.as_view()),
        name='ucomponent-list'
    ),
    path(
        'components/<slug:user>/list/',
        login_required(ComponentIndex.as_view()),
        name='ucomponent-list'
    ),
    # path(
    #     'ucs/create/<slug:user>/',
    #     ComponentCreate.as_view(),
    #     name='ucomponent-add'
    # ),
    path(
        'components/add/',
        login_required(ComponentCreate.as_view()),
        name='ucomponent-add'
    ),
    path(
        'components/<path:token>/update/',
        login_required(ComponentUpdate.as_view()),
        name='ucomponent-update'
    ),
    path(
        'components/<path:token>/delete/',
        login_required(ComponentDelete.as_view()),
        name='ucomponent-delete'
    ),

    path(
        'components/<path:token>/list/',
        ContentIndex.as_view(),
        name='ucontent-list'
    ),
    path(
        'components/<path:token>/export/',
        ContentIndex.as_view(scope="export"),
        name='ucontent-export'
    ),
    path(
        'components/<path:token>/add/<slug:type>/',
        ContentAdd.as_view(),
        name='ucontent-add'
    ),
    path(
        'content/<path:token>/delete/',
        ContentDelete.as_view(),
        name='ucontent-delete'
    ),
    path(
        'content/<path:token>/<slug:access>/',
        ContentAccess.as_view(),
        name='ucontent-access'
    ),
    path(
        'token/<path:token>/delete/',
        login_required(TokenDelete.as_view()),
        name='token-delete'
    ),
    path(
        'token/delete-request/',
        TokenDeletionRequest.as_view(),
        name='token-delete-request'
    ),
    path(
        'token/renew/',
        TokenRenewal.as_view(),
        name='token-renew'
    ),
    path(
        'components/',
        ComponentPublicIndex.as_view(
            is_home=False
        ),
        name='ucomponent-listpublic'
    ),
]



# Generated by Django 2.1.5 on 2019-01-17 12:53

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_tags', '0003_auto_20180815_1205'),
    ]

    operations = [
        migrations.AddField(
            model_name='taglayout',
            name='unique',
            field=models.BooleanField(blank=True, default=False),
        ),
        migrations.AlterField(
            model_name='taglayout',
            name='layout',
            field=jsonfield.fields.JSONField(default=list, help_text='Field list in JSON format'),
        ),
    ]



__all__ = ["CreateEntryForm"]

import tempfile
import shutil

from django import forms
from django.forms import widgets
from django.utils.translation import gettext_lazy as _
from django.conf import settings


from spkcspider.apps.spider.helpers import merge_get_url, get_settings_func
from .models import VerifySourceObject

_source_url_help = _(
    "Url to content or content list to verify"
)

_source_file_help = _(
    "File with data to verify"
)


class CreateEntryForm(forms.Form):
    url = forms.URLField(help_text=_source_url_help)
    dvfile = forms.FileField(
        required=False, max_length=settings.VERIFIER_MAX_SIZE_DIRECT_ACCEPTED
    )
    # MAX_FILE_SIZE = forms.CharField(
    #    disabled=True, widget=forms.HiddenInput(), required=False,
    #    initial=settings.VERIFIER_MAX_SIZE_ACCEPTED
    # )

    def __init__(self, instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if settings.VERIFIER_MAX_SIZE_DIRECT_ACCEPTED > 0:
            self.fields["dvfile"].help_text = _source_file_help
            self.fields["url"].required = False
        else:
            self.fields["dvfile"].disabled = True
            self.fields["dvfile"].widget = widgets.HiddenInput()

    def clean(self):
        ret = super().clean()
        if not ret.get("url", None) and not ret.get("dvfile", None):
            raise forms.ValidationError(
                _('Require either url or dvfile'),
                code="missing_parameter"
            )
            return ret
        if ret.get("url", None):
            self.cleaned_data["url"] = merge_get_url(
                self.cleaned_data["url"], raw="embed"
            )
            url = self.cleaned_data["url"]
            if not get_settings_func(
                "SPIDER_URL_VALIDATOR",
                "spkcspider.apps.spider.functions.validate_url_default"
            )(url):
                self.add_error(
                    "url", forms.ValidationError(
                        _('Insecure url: %(url)s'),
                        params={"url": url},
                        code="insecure_url"
                    )
                )
                return ret
        return ret

    def save(self):
        if self.cleaned_data.get("url", None):
            split = self.cleaned_data["url"].split("?", 1)
            return VerifySourceObject.objects.update_or_create(
                url=split[0], defaults={"get_params": split[1]}
            )[0].id
        else:
            f = tempfile.mkstemp()
            shutil.copyfileobj(self.cleaned_data["dvfile"].file, f)
            return f.name, self.cleaned_data["dvfile"].size



# Generated by Django 2.1.5 on 2019-01-09 21:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0007_linkcontent_push'),
    ]

    operations = [
        migrations.AddField(
            model_name='authtoken',
            name='referrer',
            field=models.URLField(blank=True, max_length=400, null=True),
        ),
    ]



"""
Protections
namespace: spider_base

"""

__all__ = ["Protection", "AssignedProtection", "AuthToken"]

import logging

from django.db import models
from django.conf import settings
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.views.decorators.debug import sensitive_variables

from jsonfield import JSONField

from ..constants import (
    MAX_TOKEN_B64_SIZE, hex_size_of_bigid, TokenCreationError
)
from ..helpers import create_b64_id_token, validator_token
from ..protections import installed_protections
from ..constants import ProtectionType, ProtectionResult

logger = logging.getLogger(__name__)


_striptoken = getattr(settings, "TOKEN_SIZE", 30)*4//3
# show 1/3 of token
_striptoken = _striptoken-_striptoken//3


class ProtectionManager(models.Manager):
    def invalid(self):
        return self.get_queryset().exclude(code__in=installed_protections)

    def valid(self):
        return self.get_queryset().filter(code__in=installed_protections)


# don't confuse with Protection objects used with add_protection
# this is pure DB
class Protection(models.Model):
    objects = ProtectionManager()
    # autogenerated, no choices required
    code = models.SlugField(max_length=10, primary_key=True, db_index=False)
    # protection abilities/requirements
    ptype = models.CharField(
        max_length=10, default=ProtectionType.authentication.value
    )

    @property
    def installed_class(self):
        return installed_protections[self.code]

    def __str__(self):
        return self.localize_name()

    def __repr__(self):
        return "<Protection: %s>" % self.__str__()

    def localize_name(self):
        if self.code not in installed_protections:
            return self.code
        return self.installed_class.localize_name(self.code)

    def auth_localize_name(self):
        if self.code not in installed_protections:
            return self.code
        return self.installed_class.auth_localize_name(self.code)

    @sensitive_variables("kwargs")
    def auth(self, request, obj=None, **kwargs):
        # never ever allow authentication if not active
        if obj and not obj.active:
            return False
        if self.code not in installed_protections:
            return False
        return self.installed_class.auth(
            obj=obj, request=request, **kwargs.copy()
        )

    @classmethod
    @sensitive_variables("kwargs")
    def auth_query(cls, request, query, required_passes=1, **kwargs):
        initial_required_passes = required_passes
        ret = []
        max_result = 0
        for item in query:
            obj = None
            _instant_fail = False
            if hasattr(item, "protection"):  # is AssignedProtection
                item, obj = item.protection, item
                _instant_fail = obj.instant_fail
            # would be surprising if auth fails with required_passes == 0
            # achievable by required_passes = amount of protections
            if initial_required_passes == 0:
                _instant_fail = False
            result = item.auth(
                request=request, obj=obj, query=query,
                required_passes=initial_required_passes, **kwargs
            )
            if _instant_fail:  # instant_fail does not reduce required_passes
                if type(result) is not int:  # False or form
                    # set limit unreachable
                    required_passes = len(query)
                else:
                    if result > max_result:
                        max_result = result
            elif type(result) is int:
                required_passes -= 1
                if result > max_result:
                    max_result = result
            if result is not False:  # False will be not rendered
                ret.append(ProtectionResult(result, item))
        # after side effects like raise Http404
        if (
                request.GET.get("protection", "") == "false" and
                initial_required_passes > 0
           ):
            return False
        # don't require lower limit this way
        #   against timing attacks
        if required_passes <= 0:
            return max_result
        return ret

    @classmethod
    @sensitive_variables("kwargs")
    def authall(cls, request, required_passes=1,
                ptype=ProtectionType.authentication.value,
                protection_codes=None, **kwargs):
        """
            Usage: e.g. prerendering for login fields, because
            no assigned object is available there is no config
        """
        query = cls.objects.filter(ptype__contains=ptype)

        # before protection_codes, for not allowing users
        # to manipulate required passes
        if required_passes > 0:
            # required_passes 1 and no protection means: login or token only
            required_passes = max(min(required_passes, len(query)), 1)
        else:
            query = query.filter(
                ptype__contains=ProtectionType.side_effects.value
            )

        if protection_codes:
            query = query.filter(
                code__in=protection_codes
            )
        return cls.auth_query(
            request, query.order_by("code"), required_passes=required_passes,
            ptype=ptype
        )

    def get_form(self, prefix=None, **kwargs):
        if prefix:
            protection_prefix = "{}_protections_{{}}".format(prefix)
        else:
            protection_prefix = "protections_{}"
        return self.installed_class(
            protection=self, prefix=protection_prefix.format(self.code),
            **kwargs
        )

    def render_raw(self, result):
        return {self.code: self.installed_class.render_raw(result)}

    @classmethod
    def get_forms(cls, ptype=None, **kwargs):
        protections = cls.objects.valid()
        if ptype:
            protections = protections.filter(ptype__contains=ptype)
        else:
            ptype = ""
        return map(lambda x: x.get_form(ptype=ptype, **kwargs), protections)


def get_limit_choices_assigned_protection():
    # django cannot serialize static, classmethods
    # possible?????
    index = models.Q(usercomponent__strength=10)
    restriction = models.Q(
        ~index, ptype__contains=ProtectionType.access_control.value
    )
    restriction |= models.Q(
        index, ptype__contains=ProtectionType.authentication.value
    )
    return models.Q(code__in=Protection.objects.valid()) & restriction


class AssignedProtection(models.Model):
    id = models.BigAutoField(primary_key=True)
    # fix linter warning
    objects = models.Manager()
    protection = models.ForeignKey(
        Protection, on_delete=models.CASCADE, related_name="assigned",
        limit_choices_to=get_limit_choices_assigned_protection
    )
    usercomponent = models.ForeignKey(
        "spider_base.UserComponent", related_name="protections",
        on_delete=models.CASCADE, editable=False
    )
    # data for protection
    data = JSONField(default=dict, null=False)
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True, editable=False)
    active = models.BooleanField(default=True)
    instant_fail = models.BooleanField(
        default=False,
        help_text=_("Auth fails if test fails, stronger than required_passes\n"
                    "Works even if required_passes=0\n"
                    "Does not contribute to required_passes, "
                    "ideal for side effects"
                    )
    )

    class Meta:
        unique_together = [("protection", "usercomponent")]

    def __str__(self):
        return "%s -> %s" % (
            self.usercomponent, self.protection.localize_name()
        )

    def __repr__(self):
        return "<Assigned: %s>" % (
            self.__str__()
        )

    @classmethod
    def authall(cls, request, usercomponent,
                ptype=ProtectionType.access_control.value,
                protection_codes=None, **kwargs):
        query = cls.objects.filter(
            protection__ptype__contains=ptype, active=True,
            usercomponent=usercomponent
        )
        # before protection_codes, for not allowing users
        # to manipulate required passes
        if usercomponent.required_passes > 0:
            required_passes = max(
                min(
                    usercomponent.required_passes,
                    len(query.exclude(instant_fail=True))
                ), 1
            )
        elif ptype == ProtectionType.authentication.value:
            # enforce a minimum of required_passes, if auth (e.g. index)
            required_passes = 1
        else:
            required_passes = 0
            # only protections with side effects
            query = query.filter(
                protection__ptype__contains=ProtectionType.side_effects.value
            )

        if protection_codes:
            query = query.filter(
                protection__code__in=protection_codes
            )

        return Protection.auth_query(
            request, query.order_by("protection__code"),
            required_passes=required_passes, ptype=ptype
        )

    @property
    def user(self):
        return self.usercomponent.user


class AuthToken(models.Model):
    id = models.BigAutoField(primary_key=True, editable=False)
    usercomponent = models.ForeignKey(
        "spider_base.UserComponent", on_delete=models.CASCADE,
        related_name="authtokens"
    )
    # -1=false,0=usercomponent,1-...=anchor
    persist = models.BigIntegerField(blank=True, default=-1, db_index=True)
    # brute force protection
    #  16 = usercomponent.id in hexadecimal
    #  +2 for seperators
    token = models.CharField(
        max_length=MAX_TOKEN_B64_SIZE+hex_size_of_bigid+2,
        db_index=True, unique=True,
        validators=[
            validator_token
        ]
    )
    referrer = models.ForeignKey(
        "spider_base.ReferrerObject", on_delete=models.CASCADE,
        related_name="tokens", blank=True, null=True
    )
    created_by_special_user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
        related_name="+", blank=True, null=True
    )
    session_key = models.CharField(max_length=40, null=True)
    extra = JSONField(default=dict, blank=True)
    created = models.DateTimeField(auto_now_add=True, editable=False)

    def __str__(self):
        return "{}...".format(self.token[:-_striptoken])

    def create_auth_token(self):
        self.token = create_b64_id_token(
            self.usercomponent.id, "_", getattr(settings, "TOKEN_SIZE", 30)
        )

    def save(self, *args, **kwargs):
        for i in range(0, 1000):
            if i >= 999:
                raise TokenCreationError(
                    'A possible infinite loop was detected'
                )
            self.create_auth_token()
            try:
                self.validate_unique()
                break
            except ValidationError:
                pass
        super().save(*args, **kwargs)



# Generated by Django 2.1 on 2018-08-21 00:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0003_auto_20180820_1409'),
    ]

    operations = [
        migrations.AlterField(
            model_name='textfilet',
            name='editable_from',
            field=models.ManyToManyField(blank=True, help_text='Allow editing from selected components.', related_name='_textfilet_editable_from_+', to='spider_base.UserComponent'),
        ),
    ]



# Generated by Django 2.1 on 2018-08-10 00:44

from django.db import migrations, models
import spkcspider.apps.spider_filets.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FileFilet',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('add', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=255)),
                ('file', models.FileField(upload_to=spkcspider.apps.spider_filets.models.get_file_path)),
            ],
            options={
                'abstract': False,
                'default_permissions': [],
            },
        ),
        migrations.CreateModel(
            name='TextFilet',
            fields=[
                ('id', models.BigAutoField(editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
                ('text', models.TextField(default='')),
            ],
            options={
                'abstract': False,
                'default_permissions': [],
            },
        ),
    ]




__all__ = (
    "add_content", "installed_contents", "BaseContent"
)
import logging
from urllib.parse import urljoin
from functools import lru_cache

from django.utils.html import escape
from django.apps import apps as django_apps
from django.db import models, transaction
from django.utils.translation import gettext
from django.template.loader import render_to_string
from django.core.files.base import File
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.http import Http404
from django.db.utils import IntegrityError
from django.middleware.csrf import CsrfViewMiddleware
from django.contrib.contenttypes.fields import GenericRelation
from django.http.response import HttpResponseBase, HttpResponse
from django.conf import settings
from django.utils.translation import pgettext
from django.contrib import messages
from django.views.decorators.csrf import csrf_exempt

from rdflib import Literal, Graph, BNode, URIRef, XSD

from .constants import VariantType, spkcgraph, ActionUrl
from .conf import get_anchor_domain
from .serializing import paginate_stream, serialize_stream
from .helpers import (
    get_settings_func, add_property, create_b64_id_token, merge_get_url
)
from .templatetags.spider_rdf import literalize

installed_contents = {}

# don't spam set objects
_empty_set = frozenset()

default_abilities = frozenset(
    ("add", "view", "update", "export", "list", "raw", "raw_update")
)

# never use these names
forbidden_names = ["Content", "UserComponent"]

# updated attributes of ContentVariant
_attribute_list = ["name", "ctype", "strength"]


def add_content(klass):
    code = klass._meta.model_name
    if code in installed_contents:
        raise Exception("Duplicate content")
    if "{}.{}".format(klass.__module__, klass.__qualname__) not in getattr(
        settings, "SPIDER_BLACKLISTED_MODULES", _empty_set
    ):
        installed_contents[code] = klass
    return klass


def initialize_content_models(apps=None):
    if not apps:
        apps = django_apps
    ContentVariant = apps.get_model("spider_base", "ContentVariant")
    all_content = models.Q()
    for code, val in installed_contents.items():
        appearances = val.appearances
        if callable(appearances):
            appearances = appearances()

        # update name if only one name exists
        update = False
        if len(appearances) == 1:
            update = True

        for dic in appearances:
            require_save = False
            assert dic["name"] not in forbidden_names, \
                "Forbidden content name: %" % dic["name"]
            try:
                with transaction.atomic():
                    if update:
                        variant = ContentVariant.objects.get_or_create(
                            defaults=dic, code=code
                        )[0]
                    else:
                        variant = ContentVariant.objects.get_or_create(
                            defaults=dic, code=code, name=dic["name"]
                        )[0]

            except IntegrityError:
                # renamed model = code changed
                variant = ContentVariant.objects.get(name=dic["name"])
                variant.code = code
                require_save = True

            for key in _attribute_list:
                val = dic.get(
                    key, variant._meta.get_field(key).get_default()
                )
                if getattr(variant, key) != val:
                    setattr(variant, key, val)
                    require_save = True
            if require_save:
                variant.save()
            all_content |= models.Q(name=dic["name"], code=code)
    invalid_models = ContentVariant.objects.exclude(all_content)
    if invalid_models.exists():
        print("Invalid content, please update or remove them:",
              ["\"{}\":{}".format(t.code, t.name) for t in invalid_models])


class BaseContent(models.Model):
    # consider not writing admin wrapper for (sensitive) inherited content
    # this way content could be protected to be only visible to admin, user
    # and legitimated users (if not index)

    # iterable or callable returning iterable containing dicts
    # required keys: name
    # optional: strength (default=0), ctype (default="")
    # max name length is 50 chars
    # max ctype length is 10 chars (10 attributes)

    # use case: addons for usercontent, e.g. dependencies on external libraries
    # use case: model with different abilities
    appearances = None

    id = models.BigAutoField(primary_key=True, editable=False)
    # every content can specify its own deletion period
    deletion_period = getattr(
        settings, "SPIDER_CONTENTS_DEFAULT_DELETION_PERIOD", None
    )
    # if created associated is None (will be set later)
    # use usercomponent in form instead
    associated_rel = GenericRelation("spider_base.AssignedContent")
    _associated_tmp = None

    # user can set name
    # if set to "force" name will be enforced
    expose_name = True
    # user can set description
    expose_description = False

    associated_errors = None

    _content_is_cleaned = False

    @property
    def associated(self):
        if self._associated_tmp:
            return self._associated_tmp
        return self.associated_rel.first()

    class Meta:
        abstract = True
        default_permissions = []

    @classmethod
    def static_create(
        cls, *, associated=None, associated_kwargs=None, content_kwargs=None,
        token_size=None
    ):
        if content_kwargs:
            ob = cls(**content_kwargs)
        else:
            ob = cls()
        if associated:
            ob._associated_tmp = associated
        else:
            from .models import AssignedContent
            ob = cls()
            if not associated_kwargs:
                associated_kwargs = {}
            ob._associated_tmp = AssignedContent(**associated_kwargs)
        ob._associated_tmp.content = ob
        if token_size is not None:
            ob._associated_tmp.token_generate_new_size = token_size
        return ob

    @classmethod
    def localize_name(cls, name):
        return pgettext("content name", name)

    def __str__(self):
        if not self.id:
            return self.localize_name(self.associated.ctype.name)
        else:
            return self.associated.name

    def __repr__(self):
        return "<Content: ({}: {})>".format(
            self.associated.usercomponent.username, self.__str__()
        )

    def get_size(self):
        # 255 = length name no matter what encoding
        s = 255
        if self.expose_description:
            s += len(self.associated.description)
        return s

    def get_priority(self):
        return 0

    @classmethod
    def feature_urls(cls, name):
        """ For implementing component features """
        return []

    @classmethod
    @lru_cache(typed=True)
    def cached_feature_urls(cls, name):
        return list(map(
            lambda x: ActionUrl(*x),
            cls.feature_urls(name)
        ))

    def get_content_name(self):
        return "{}_{}".format(
            self.localize_name(self.associated.ctype.name),
            self.associated.id
        )

    def get_content_description(self):
        return " "

    def localized_description(self):
        """ localize and perform other transforms before rendering to user """
        if not self.expose_description:
            return self.associated.description
        return gettext(self.associated.description)

    def get_strength(self):
        """ get required strength """
        return self.associated.ctype.strength

    def get_strength_link(self):
        """ get required strength for links """
        return self.get_strength()

    def get_abilities(self, context):
        """ Override for declaring content extra abilities """
        return set()

    def get_instance_form(self, context):
        return self

    def get_form_kwargs(
        self, request, instance=None, disable_data=False, **kwargs
    ):
        """Return the keyword arguments for instantiating the form."""
        fkwargs = {}
        if instance is None:
            fkwargs["instance"] = self.get_instance_form(kwargs)
        elif instance:
            fkwargs["instance"] = instance

        if not disable_data and request.method in ('POST', 'PUT'):
            fkwargs.update({
                'data': request.POST,
                'files': request.FILES,
            })
        return fkwargs

    # default scopes:
    #  public accessible:
    #   raw: view with raw parameter
    #   list: raw but embedded in content list
    #   view: view of conten without raw parameter
    #  only accessible by owner and staff:
    #   add: create view of content
    #   update: update view of content
    #   raw_update: special update mode for machines or really fancy updates
    #   export: like raw just optimized for export: e.g. no dereferencing
    def get_form(self, scope):
        raise NotImplementedError

    def get_template_name(self, scope):
        if scope in ["add", "update"]:
            return 'spider_base/partials/base_form.html'
        return 'spider_base/view_form.html'

    def update_used_space(self, size_diff):
        if size_diff == 0:
            return
        f = "local"
        if VariantType.component_feature.value in self.associated.ctype.ctype:
            f = "remote"
        with transaction.atomic():
            self.associated.usercomponent.user_info.update_with_quota(
                    size_diff, f
                )
            self.associated.usercomponent.user_info.save(
                update_fields=[
                    "used_space_local", "used_space_remote"
                ]
            )

    def render_form(self, scope, **kwargs):
        _ = gettext
        if scope == "add":
            kwargs["form_empty_message"] = _("<b>No User Input required</b>")
            old_size = 0
        else:
            old_size = self.get_size()
        parent_form = kwargs.get("form", None)
        kwargs["form"] = self.get_form(scope)(
            **self.get_form_kwargs(
                scope=scope,
                **kwargs
            )
        )
        if kwargs["form"].is_valid():
            instance = kwargs["form"].save(False)
            try:
                self.update_used_space(
                    instance.get_size() - old_size
                )
            except ValidationError as exc:
                kwargs["form"].add_error(None, exc)
                messages.error(
                    kwargs["request"], _('Space exhausted')
                )
        if kwargs["form"].is_valid():
            instance.save()
            kwargs["form"].save_m2m()
            messages.success(
                kwargs["request"], _('Content updated')
            )
            kwargs["form"] = self.get_form(scope)(
                **self.get_form_kwargs(
                    scope=scope,
                    instance=instance,
                    **kwargs
                )
            )

        elif parent_form:
            if len(kwargs["form"].errors.get(NON_FIELD_ERRORS, [])) > 0:
                parent_form.add_error(
                    None, kwargs["form"].errors.pop(NON_FIELD_ERRORS)
                )
            if self.associated_errors:
                parent_form.add_error(
                    None, self.associated_errors
                )
        return (
            render_to_string(
                self.get_template_name(scope),
                request=kwargs["request"], context=kwargs
            ),
            kwargs["form"].media
        )

    def get_absolute_url(self, scope="view"):
        return self.associated.get_absolute_url(scope)

    def get_primary_anchor(self, graph, context):
        p = self.associated.usercomponent.primary_anchor
        if p:
            return urljoin(
                get_anchor_domain(),
                p.get_absolute_url()
            )

        if not p and self.associated.usercomponent.public:
            return urljoin(
                get_anchor_domain(),
                self.associated.usercomponent.get_absolute_url()
            )
        return urljoin(
            get_anchor_domain(), context["request"].path
        )

    def get_references(self):
        return []

    def map_data(self, name, field, data, graph, context):
        if isinstance(data, File):
            return get_settings_func(
                "SPIDER_FILE_EMBED_FUNC",
                "spkcspider.apps.spider.functions.embed_file_default"
            )(name, data, self, context)
        return literalize(data, field)

    def serialize(self, graph, ref_content, context):
        form = self.get_form(context["scope"])(
            **self.get_form_kwargs(
                disable_data=True,
                **context
            )
        )
        graph.add((
            ref_content,
            spkcgraph["type"],
            Literal(self.associated.getlist("type", 1)[0])
        ))
        if "abilities" not in context:
            context["abilities"] = set(self.get_abilities(context))

        for ability in context["abilities"]:
            assert(ability not in default_abilities)

            graph.add((
                ref_content,
                spkcgraph["ability:name"],
                Literal(ability, datatype=XSD.string)
            ))

        for name, field in form.fields.items():
            raw_value = form.initial.get(name, None)
            try:
                value = field.to_python(raw_value)
            except Exception as exc:
                # user can corrupt tags, so just debug
                level = logging.WARNING
                if getattr(form, "layout_generating_form", False):
                    level = logging.DEBUG
                logging.log(
                    level,
                    "Corrupted field: %s, form: %s, error: %s",
                    name, form, exc
                )
                continue
            value_node = BNode()
            hashable = getattr(field, "hashable", False)

            graph.add((
                ref_content,
                spkcgraph["properties"],
                value_node
            ))
            graph.add((
                value_node,
                spkcgraph["hashable"],
                Literal(hashable)
            ))
            graph.add((
                value_node,
                spkcgraph["name"],
                Literal(name, datatype=XSD.string)
            ))
            graph.add((
                value_node,
                spkcgraph["fieldname"],
                Literal(form.add_prefix(name), datatype=XSD.string)
            ))

            if not isinstance(value, (list, tuple, models.QuerySet)):
                value = [value]

            for i in value:
                graph.add((
                    value_node,
                    spkcgraph["value"],
                    self.map_data(name, field, i, graph, context)
                ))

    def render_serialize(self, **kwargs):
        from .models import AssignedContent
        # ** creates copy of dict, so it is safe to overwrite kwargs here

        session_dict = {
            "request": kwargs["request"],
            "context": kwargs,
            "scope": kwargs["scope"],
            "hostpart": kwargs["hostpart"],
            "ac_namespace": spkcgraph["contents"],
            "sourceref": URIRef(urljoin(
                kwargs["hostpart"], kwargs["request"].path
            ))
        }

        g = Graph()
        g.namespace_manager.bind("spkc", spkcgraph, replace=True)

        p = paginate_stream(
            AssignedContent.objects.filter(id=self.associated.id),
            getattr(
                settings, "SPIDER_SERIALIZED_PER_PAGE",
                settings.SPIDER_OBJECTS_PER_PAGE
            ),
            settings.SPIDER_MAX_EMBED_DEPTH
        )
        page = 1
        try:
            page = int(self.request.GET.get("page", "1"))
        except Exception:
            pass
        serialize_stream(
            g, p, session_dict,
            page=page,
            embed=True
        )
        if hasattr(kwargs["request"], "token_expires"):
            session_dict["expires"] = kwargs["request"].token_expires.strftime(
                "%a, %d %b %Y %H:%M:%S %z"
            )
            if page <= 1:
                add_property(
                    g, "token_expires", ob=session_dict["request"],
                    ref=session_dict["sourceref"]
                )

        if page <= 1:
            g.add((
                session_dict["sourceref"],
                spkcgraph["scope"],
                Literal(kwargs["scope"], datatype=XSD.string)
            ))

            uc = kwargs.get("source", self.associated.usercomponent)
            g.add((
                session_dict["sourceref"], spkcgraph["strength"],
                Literal(uc.strength)
            ))
            token = getattr(session_dict["request"], "auth_token", None)
            if token:
                token = token.token
            url2 = merge_get_url(str(session_dict["sourceref"]), token=token)
            g.add(
                (
                    session_dict["sourceref"],
                    spkcgraph["action:view"],
                    URIRef(url2)
                )
            )
            if kwargs["referrer"]:
                g.add((
                    session_dict["sourceref"],
                    spkcgraph["referrer"],
                    Literal(kwargs["referrer"], datatype=XSD.anyURI)
                ))
            if kwargs["token_strength"]:
                add_property(
                    g, "token_strength", ref=session_dict["sourceref"],
                    literal=kwargs["token_strength"]
                )
            add_property(
                g, "intentions", ref=session_dict["sourceref"],
                literal=kwargs["intentions"], datatype=XSD.string,
                iterate=True
            )

        ret = HttpResponse(
            g.serialize(format="turtle"),
            content_type="text/turtle;charset=utf-8"
        )

        if session_dict.get("expires", None):
            ret['X-Token-Expires'] = session_dict["expires"]
        # allow cors requests for raw
        ret["Access-Control-Allow-Origin"] = "*"
        return ret

    def access_view(self, **kwargs):
        _ = gettext

        kwargs["form"] = self.get_form("view")(
            **self.get_form_kwargs(disable_data=True, **kwargs)
        )
        kwargs.setdefault("legend", escape(_("View")))
        # not required: done by access template
        # kwargs.setdefault(
        #    "add_spkc_types",
        #    [
        #        self.associated.getlist("type", 1)[0],
        #        "Content"
        #    ]
        # )
        return (
            render_to_string(
                self.get_template_name(kwargs["scope"]),
                request=kwargs["request"],
                context=kwargs
            ),
            kwargs["form"].media
        )

    def access_add(self, **kwargs):
        _ = gettext
        kwargs.setdefault(
            "legend",
            escape(_("Add \"%s\"") % self.__str__())
        )
        # not visible by default
        kwargs.setdefault("confirm", _("Create"))
        # prevent second button
        kwargs.setdefault("inner_form", True)
        return self.render_form(**kwargs)

    def access_update(self, **kwargs):
        _ = gettext
        kwargs.setdefault(
            "legend",
            escape(_("Update \"%s\"") % self.__str__())
        )
        # not visible by default
        kwargs.setdefault("confirm", _("Update"))
        # prevent second button
        kwargs.setdefault("inner_form", True)
        return self.render_form(**kwargs)

    def access_export(self, **kwargs):
        kwargs["scope"] = "export"
        return self.render_serialize(**kwargs)

    def access_raw(self, **kwargs):
        kwargs["scope"] = "raw"
        return self.render_serialize(**kwargs)

    @csrf_exempt
    def access_default(self, **kwargs):
        raise Http404()

    access_raw_update = access_default

    def access(self, context):
        # context is updated and used outside!!
        # so make sure that func gets only a copy (**)
        context["abilities"] = set(self.get_abilities(context))
        context.setdefault("extra_outer_forms", [])
        func = self.access_default
        if context["scope"] == "view" and "raw" in context["request"].GET:
            func = self.access_raw
        elif context["scope"] in default_abilities:
            func = getattr(self, "access_{}".format(context["scope"]))
        elif context["scope"] in context["abilities"]:
            func = getattr(self, "access_{}".format(context["scope"]))

        # check csrf tokens manually
        if not getattr(func, "csrf_exempt", False):
            csrferror = CsrfViewMiddleware().process_view(
                context["request"], None, (), {}
            )
            if csrferror is not None:
                # csrferror is HttpResponse
                return csrferror
        ret = func(**context)
        if context["scope"] == "update":
            # update function should never return HttpResponse
            #  (except csrf error)
            assert(not isinstance(ret, HttpResponseBase))
        return ret

    def get_info(self, unique=None, unlisted=None):
        # unique=None, feature=None shortcuts for get_info overwrites
        # passing down these parameters not neccessary
        if unique is None:
            unique = (
                VariantType.unique.value in self.associated.ctype.ctype
            )
        if unlisted is None:
            unlisted = (
                VariantType.component_feature.value in
                self.associated.ctype.ctype or
                VariantType.content_feature.value in
                self.associated.ctype.ctype
            )

        anchortag = ""
        if VariantType.anchor.value in self.associated.ctype.ctype:
            anchortag = "anchor\x1e"

        idtag = "primary\x1e"
        # simulates beeing not unique, by adding id
        if not unique:
            idtag = "id=\x1e"  # placeholder
            if getattr(self.associated, "id", None):
                idtag = "id={}\x1e".format(self.associated.id)
        unlistedtag = ""
        if unlisted:
            unlistedtag = "unlisted\x1e"
        return "\x1etype={}\x1e{}{}{}".format(
            self.associated.ctype.name,
            idtag,
            anchortag,
            unlistedtag
        )

    def full_clean(self, **kwargs):
        # checked with clean
        kwargs.setdefault("exclude", []).append("associated_rel")
        return super().full_clean(**kwargs)

    def clean(self):
        _ = gettext
        if self._associated_tmp:
            self._associated_tmp.content = self
        assignedcontent = self.associated
        assignedcontent.info = self.get_info()
        if getattr(self, "id", None):
            if not self.expose_name or not assignedcontent.name:
                assignedcontent.name = self.get_content_name()
            if not self.expose_description or not assignedcontent.description:
                assignedcontent.description = self.get_content_description()
        assignedcontent.priority = self.get_priority()
        assignedcontent.strength = self.get_strength()
        assignedcontent.strength_link = self.get_strength_link()
        try:
            assignedcontent.full_clean(exclude=["content"])
        except ValidationError as exc:
            self.associated_errors = exc
        # persist AssignedContent for saving
        self._associated_tmp = assignedcontent
        if self.associated_errors:
            raise ValidationError(
                _('AssignedContent validation failed'),
                code="assigned_content"
            )
        self._content_is_cleaned = True

    def save(self, *args, **kwargs):
        if settings.DEBUG:
            assert self._content_is_cleaned, "try to save uncleaned content"
        super().save(*args, **kwargs)
        assignedcontent = self.associated
        if not assignedcontent.content:
            # add requires this
            assignedcontent.content = self
            assignedcontent.info = self.get_info()
            assignedcontent.strength = self.get_strength()
            assignedcontent.strength_link = self.get_strength_link()
        created = False
        if not getattr(assignedcontent, "id", None):
            created = True
        # update info and set content
        assignedcontent.save()
        if created:
            to_save = set()
            # add id to info
            if "\x1eprimary\x1e" not in assignedcontent.info:
                assignedcontent.info = assignedcontent.info.replace(
                    "\x1eid=\x1e", "\x1eid={}\x1e".format(
                        assignedcontent.id
                    ), 1
                )
                to_save.add("info")
            if (
                assignedcontent.token_generate_new_size is not None and
                not assignedcontent.token
            ):
                assignedcontent.token_generate_new_size = \
                    getattr(settings, "TOKEN_SIZE", 30)
            # set token
            if assignedcontent.token_generate_new_size is not None:
                if assignedcontent.token:
                    print(
                        "Old nonce for Content id:", assignedcontent.id,
                        "is", assignedcontent.token
                    )
                assignedcontent.token = create_b64_id_token(
                    assignedcontent.id,
                    "/",
                    assignedcontent.token_generate_new_size
                )
                assignedcontent.token_generate_new_size = None
                to_save.add("token")
            if not self.expose_name or not assignedcontent.name:
                assignedcontent.name = self.get_content_name()
                to_save.add("name")
            if not self.expose_description or not assignedcontent.description:
                assignedcontent.description = self.get_content_description()
                to_save.add("description")
            # second save required
            if to_save:
                assignedcontent.save(update_fields=to_save)
        # needs id first
        s = set(assignedcontent.attached_contents.all())
        s.update(self.get_references())
        assignedcontent.references.set(s)
        # delete saved errors
        self.associated_errors = None
        # require cleaning again
        self._content_is_cleaned = False



# Generated by Django 2.1.5 on 2019-01-20 12:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0012_auto_20190118_1017'),
    ]

    operations = [
        migrations.AddField(
            model_name='assignedcontent',
            name='persist_token',
            field=models.ForeignKey(blank=True, limit_choices_to={'persist__gte': 0}, null=True, on_delete=django.db.models.deletion.CASCADE, to='spider_base.AuthToken'),
        ),
    ]




from django.conf import settings
from django.urls import path, reverse_lazy

from django.contrib.auth.views import (
    PasswordChangeView, PasswordChangeDoneView
)
from django.views.generic.base import RedirectView, TemplateView

from .views import SignupView, UserUpdateView, UserLoginView, UserLogoutView


app_name = "auth"

# no recovery, only authentication

urlpatterns = [
    path('login/', UserLoginView.as_view(), name='login'),
    path('logout/', UserLogoutView.as_view(), name='logout'),
    path(
        'password_change/',
        PasswordChangeView.as_view(
            success_url=reverse_lazy("auth:password_change_done")
        ), name='password_change'
    ),
    path(
        'password_change/done/',
        PasswordChangeDoneView.as_view(), name='password_change_done'
    ),
    path(
        'thanks/',
        TemplateView.as_view(template_name='registration/thanks.html'),
        name="signup_thanks"
    ),
    path('profile/', UserUpdateView.as_view(), name="profile"),
    path(
        '',
        RedirectView.as_view(
            url=reverse_lazy('auth:profile'),
            permanent=True
        )
    ),
]
if getattr(settings, "OPEN_FOR_REGISTRATION", False):
    urlpatterns.append(
        path('signup/', SignupView.as_view(), name="signup")
    )



""" Content Views """

__all__ = (
    "ContentIndex", "ContentAdd", "ContentAccess", "ContentDelete"
)

from datetime import timedelta

from django.views.generic.edit import DeleteView, UpdateView, CreateView
from django.views.generic.list import ListView
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.http.response import HttpResponseBase, HttpResponse
from django.db import models
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.contrib import messages
from django.utils.translation import gettext
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator

from next_prev import next_in_order, prev_in_order

from rdflib import Graph, Literal, URIRef, XSD


from ._core import UCTestMixin, EntityDeletionMixin, ReferrerMixin
from ..models import (
    AssignedContent, ContentVariant, UserComponent
)
from ..forms import UserContentForm
from ..helpers import get_settings_func, add_property, merge_get_url
from ..queryfilters import filter_contents
from ..constants import spkcgraph, VariantType
from ..serializing import paginate_stream, serialize_stream

_forbidden_scopes = frozenset(["add", "list", "raw", "delete"])


class ContentBase(UCTestMixin):
    model = AssignedContent
    scope = None
    object = None
    # use token of content object instead
    no_token_usercomponent = True

    def dispatch(self, request, *args, **kwargs):
        try:
            return super().dispatch(request, *args, **kwargs)
        except Http404:
            return get_settings_func(
                "RATELIMIT_FUNC",
                "spkcspider.apps.spider.functions.rate_limit_default"
            )(self, request)

    def get_template_names(self):
        if self.scope in ("add", "update"):
            return ['spider_base/assignedcontent_form.html']
        elif self.scope == "list":
            return ['spider_base/assignedcontent_list.html']
        else:
            return ['spider_base/assignedcontent_access.html']

    def get_ordering(self, issearching=False):
        if self.scope != "list":
            # export: also serializer, other scopes: only one object, overhead
            return None
        # ordering will happen in serializer
        if "raw" in self.request.GET:
            return None
        return ("-priority", "-modified")

    def get_context_data(self, **kwargs):
        kwargs["request"] = self.request
        kwargs["scope"] = self.scope
        kwargs["uc"] = self.usercomponent
        kwargs["enctype"] = "multipart/form-data"
        return super().get_context_data(**kwargs)

    def get_queryset(self):
        ret = self.model.objects.all()
        # skip search if user and single object
        if self.scope in ("add", "update", "raw_update"):
            return ret

        if getattr(self.request, "auth_token", None):
            idlist = \
                self.request.auth_token.extra.get("ids", [])
            searchlist = \
                self.request.auth_token.extra.get("filter", [])
        else:
            idlist = []
            searchlist = []

        if self.scope == "list":
            if "search" in self.request.POST or "id" in self.request.POST:
                searchlist += self.request.POST.getlist("search")
                idlist += self.request.POST.getlist("id")
            else:
                searchlist += self.request.GET.getlist("search")
                idlist += self.request.GET.getlist("id")
        elif self.scope not in ("add", "update", "raw_update"):
            searchlist += self.request.GET.getlist("search")

        # list only unlisted if explicity requested or export or:
        # if it has high priority (only for special users)
        # listing prioritized, unlisted content is different to the broader
        # search
        filter_unlisted = False
        if self.request.is_special_user:
            # all other scopes than list can show here _unlisted
            # this includes export
            if self.scope == "list" and "_unlisted" not in searchlist:
                filter_unlisted = 0
        else:
            filter_unlisted = True

        filter_q, counter = filter_contents(
            searchlist, idlist, filter_unlisted
        )

        order = self.get_ordering(counter > 0)
        # distinct required?
        ret = ret.filter(filter_q)
        if order:
            ret = ret.order_by(*order)
        return ret


class ContentIndex(ReferrerMixin, ContentBase, ListView):
    model = AssignedContent
    scope = "list"
    no_token_usercomponent = False

    def dispatch_extra(self, request, *args, **kwargs):
        self.allow_domain_mode = self.usercomponent.allow_domain_mode
        if "referrer" in self.request.GET:
            self.object_list = self.get_queryset()
            return self.handle_referrer()
        return None

    def get_queryset(self):
        return super().get_queryset().filter(
            usercomponent=self.usercomponent
        )

    def get_usercomponent(self):
        query = {"token": self.kwargs["token"]}
        return get_object_or_404(
            UserComponent.objects.select_related(
                "user", "user__spider_info",
            ).prefetch_related("protections"),
            **query
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if self.request.is_owner:
            # request.user is maybe anonymous
            context["content_variants"] = \
                self.usercomponent.user.spider_info.allowed_content.exclude(
                    models.Q(
                        ctype__contains=VariantType.component_feature.value
                    ) |
                    models.Q(
                        ctype__contains=VariantType.content_feature.value
                    ) |
                    models.Q(ctype__contains=VariantType.unlisted.value)
                )
            context["content_variants_used"] = \
                self.usercomponent.user.spider_info.allowed_content.filter(
                    assignedcontent__usercomponent=self.usercomponent
                ).exclude(
                    models.Q(
                        ctype__contains=VariantType.component_feature.value
                    ) |
                    models.Q(
                        ctype__contains=VariantType.content_feature.value
                    ) |
                    models.Q(ctype__contains=VariantType.unlisted.value)
                )
        context["active_features"] = self.usercomponent.features.all()
        context["is_public_view"] = self.usercomponent.public
        context["has_unlisted"] = self.usercomponent.contents.filter(
            info__contains="\x1eunlisted\x1e"
        ).exists()

        context["remotelink"] = context["spider_GET"].copy()
        context["auth_token"] = None
        if self.request.auth_token:
            context["auth_token"] = self.request.auth_token.token
        context["remotelink"] = "{}{}?{}".format(
            context["hostpart"],
            reverse("spider_base:ucontent-list", kwargs={
                "token": self.usercomponent.token
            }),
            context["remotelink"].urlencode()
        )
        return context

    def test_func(self):
        staff_perm = not self.usercomponent.is_index
        if staff_perm:
            staff_perm = "spider_base.view_usercomponent"
        # user token is tested later
        if self.has_special_access(
            user_by_login=True, user_by_token=False,
            staff=staff_perm, superuser=True
        ):
            self.request.auth_token = self.create_admin_token()
            return True
        # block view on special objects for non user and non superusers
        if self.usercomponent.is_index:
            return False
        # export is only available for user and staff with permission

        minstrength = 0
        if self.scope in ["export"]:
            minstrength = 4

        return self.test_token(minstrength)

    def get_paginate_by(self, queryset):
        if self.scope == "export" or "raw" in self.request.GET:
            return None
        return settings.SPIDER_OBJECTS_PER_PAGE

    def render_to_response(self, context):
        if context["scope"] != "export" and "raw" not in self.request.GET:
            return super().render_to_response(context)

        session_dict = {
            "request": self.request,
            "context": context,
            "scope": context["scope"],
            "uc": self.usercomponent,
            "hostpart": context["hostpart"],
            "sourceref": URIRef(context["hostpart"] + self.request.path)
        }
        g = Graph()
        g.namespace_manager.bind("spkc", spkcgraph, replace=True)

        embed = False
        if (
            context["scope"] == "export" or
            self.request.GET.get("raw", "") == "embed"
        ):
            embed = True

        if context["object_list"]:
            p = paginate_stream(
                context["object_list"],
                getattr(
                    settings, "SPIDER_SERIALIZED_PER_PAGE",
                    settings.SPIDER_OBJECTS_PER_PAGE
                ),
                settings.SPIDER_MAX_EMBED_DEPTH
            )
        else:
            # no content, pagination works here only this way
            p = paginate_stream(
                UserComponent.objects.filter(pk=self.usercomponent.pk),
                1,
                1
            )
        page = 1
        try:
            page = int(self.request.GET.get("page", "1"))
        except Exception:
            pass

        if page <= 1:
            if hasattr(self.request, "token_expires"):
                session_dict["expires"] = self.request.token_expires.strftime(
                    "%a, %d %b %Y %H:%M:%S %z"
                )
                add_property(
                    g, "token_expires", ob=session_dict["request"],
                    ref=session_dict["sourceref"]
                )
            g.add((
                session_dict["sourceref"],
                spkcgraph["scope"],
                Literal(context["scope"], datatype=XSD.string)
            ))
            g.add((
                session_dict["sourceref"],
                spkcgraph["strength"],
                Literal(self.usercomponent.strength, datatype=XSD.integer)
            ))
            if context["referrer"]:
                g.add((
                    session_dict["sourceref"],
                    spkcgraph["referrer"],
                    Literal(context["referrer"], datatype=XSD.anyURI)
                ))
            token = getattr(session_dict["request"], "auth_token", None)
            if token:
                token = token.token
            g.add(
                (
                    session_dict["sourceref"],
                    spkcgraph["action:view"],
                    Literal(
                        merge_get_url(
                            str(session_dict["sourceref"]), token=token
                        ),
                        datatype=XSD.anyURI
                    )
                )
            )
            if context["token_strength"]:
                add_property(
                    g, "token_strength", ref=session_dict["sourceref"],
                    literal=context["token_strength"], datatype=XSD.integer
                )
            add_property(
                g, "intentions", ref=session_dict["sourceref"],
                literal=context["intentions"], datatype=XSD.string,
                iterate=True
            )

        serialize_stream(
            g, p, session_dict,
            page=page,
            embed=embed
        )

        ret = HttpResponse(
            g.serialize(format="turtle"),
            content_type="text/turtle;charset=utf-8"
        )

        if session_dict.get("expires", None):
            ret['X-Token-Expires'] = session_dict["expires"]
        # allow cors requests for raw
        ret["Access-Control-Allow-Origin"] = "*"
        return ret


class ContentAdd(ContentBase, CreateView):
    scope = "add"
    model = ContentVariant

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        return self.render_to_response(self.get_context_data())

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        return self.render_to_response(self.get_context_data())

    def form_valid(self, form):
        _ = gettext
        messages.success(
            self.request, _('Content created.')
        )
        return super().form_valid(form)

    def get_queryset(self):
        # use requesting user as base if he can add this type of content
        if self.request.user.is_authenticated:
            return self.request.user.spider_info.allowed_content.exclude(
                models.Q(ctype__contains=VariantType.component_feature.value) |
                models.Q(ctype__contains=VariantType.unlisted.value)
            )
        else:
            return self.usercomponent.user.spider_info.allowed_content.exclude(
                models.Q(ctype__contains=VariantType.component_feature.value) |
                models.Q(ctype__contains=VariantType.unlisted.value)
            )

    def test_func(self):
        # test if user and check if user is allowed to create content
        if self.has_special_access(
            user_by_login=True, user_by_token=True, superuser=False
        ):
            return True
        return False

    def get_context_data(self, **kwargs):
        kwargs["content_type"] = self.object.installed_class
        kwargs["form"] = self.get_form()
        kwargs["active_features"] = self.usercomponent.features.all()
        return super().get_context_data(**kwargs)

    def get_form(self, allow_data=True):
        assigned = self.object.installed_class.static_create(
            associated_kwargs={
                "usercomponent": self.usercomponent,
                "ctype": self.object
            }
        ).associated
        form_kwargs = {
            "instance": assigned,
            "request": self.request,
            "initial": {
                "usercomponent": self.usercomponent
            }
        }
        if allow_data and self.request.method in ('POST', 'PUT'):
            form_kwargs.update({
                'data': self.request.POST,
                # 'files': self.request.FILES,
            })
        return UserContentForm(**form_kwargs)

    def get_usercomponent(self):
        return get_object_or_404(
            UserComponent.objects.prefetch_related("protections"),
            token=self.kwargs["token"]
        )

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        qquery = models.Q(
            name=self.kwargs["type"],
            strength__lte=self.usercomponent.strength
        )
        return get_object_or_404(queryset, qquery)

    def render_to_response(self, context):
        # only true if data
        if context["form"].is_valid():
            ucontent = context["form"].save(commit=False)
        else:
            ucontent = context["form"].instance
        rendered = ucontent.content.access(context)

        # return response if content returned response
        if isinstance(rendered, HttpResponseBase):
            return rendered
        # show framed output
        assert(isinstance(rendered, (tuple, list)))
        context["content"] = rendered
        # redirect if saving worked
        if getattr(ucontent, "id", None):
            assert(ucontent.token)
            assert(ucontent.usercomponent)
            return redirect(
                'spider_base:ucontent-access',
                token=ucontent.token, access="update"
            )
        else:
            assert(not getattr(ucontent.content, "id", None))
        return super().render_to_response(context)


class ContentAccess(ReferrerMixin, ContentBase, UpdateView):
    scope = "access"
    form_class = UserContentForm
    model = AssignedContent

    def dispatch_extra(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.allow_domain_mode = (
            self.usercomponent.allow_domain_mode or
            self.object.allow_domain_mode
        )
        # done in get_queryset
        # if getattr(self.request, "auth_token", None):
        #     ids = self.request.auth_token.extra.get("ids", None)
        #     if ids is not None and self.object.id not in ids:
        #         return self.handle_no_permission()
        if "referrer" in self.request.GET:
            self.object_list = self.model.objects.filter(
                pk=self.object.pk
            )
            return self.handle_referrer()

        return None

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        _scope = kwargs["access"]
        # special scopes which should be not available as url parameter
        # raw is also deceptive because view and raw=? = raw scope
        if _scope in _forbidden_scopes:
            raise PermissionDenied("Deceptive scopes")
        self.scope = _scope
        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        context = {"form": None}
        if self.scope == "update":
            context["form"] = self.get_form()
        return self.render_to_response(self.get_context_data(**context))

    def post(self, request, *args, **kwargs):
        context = {"form": None}
        # other than update have no form
        if self.scope == "update":
            context["form"] = self.get_form()
            if context["form"].is_valid():
                self.object = context["form"].save(commit=True)
                # use correct usercomponent
                self.usercomponent = context["form"].instance.usercomponent
        return self.render_to_response(self.get_context_data(**context))

    def get_context_data(self, **kwargs):
        kwargs["is_public_view"] = (
            self.usercomponent.public and
            self.scope not in ("add", "update", "raw_update")
        )
        context = super().get_context_data(**kwargs)

        context["remotelink"] = context["spider_GET"].copy()
        context["auth_token"] = None
        if self.request.auth_token:
            context["auth_token"] = self.request.auth_token.token
        context["remotelink"] = "{}{}?{}".format(
            context["hostpart"],
            reverse("spider_base:ucontent-access", kwargs={
                "token": self.object.token,
                "access": "view"
            }),
            context["remotelink"].urlencode()
        )
        if self.scope == "update":
            context["active_features"] = self.usercomponent.features.all()
        else:
            context["active_features"] = ContentVariant.objects.filter(
                models.Q(feature_for_contents=self.object) |
                models.Q(feature_for_components=self.usercomponent)
            )
        return context

    def get_form_success_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        return {
            'initial': self.get_initial(),
            'instance': self.object,
            'request': self.request,
            'prefix': self.get_prefix()
        }

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        ret = super().get_form_kwargs()
        ret["request"] = self.request
        return ret

    def test_func(self):
        # give user and staff the ability to update Content
        # except it is index, in this case only the user can update
        # reason: admins could be tricked into malicious updates
        # for index the same reason as for add
        uncritically = self.usercomponent.name != "index"
        staff_perm = uncritically
        if staff_perm:
            staff_perm = "spider_base.view_assignedcontent"
            if self.scope in {"update", "raw_update"}:
                staff_perm = "spider_base.update_assignedcontent"
        # user token is tested later
        if self.has_special_access(
            staff=staff_perm, superuser=uncritically,
            user_by_token=False, user_by_login=True
        ):
            self.request.auth_token = self.create_admin_token()
            return True
        minstrength = 0
        if self.scope in {"update", "raw_update", "export"}:
            minstrength = 4
        return self.test_token(minstrength)

    def render_to_response(self, context):
        # context is updated and used outside!!
        rendered = self.object.content.access(context)

        if self.scope == "update":
            # token changed => path has changed
            if self.object.token != self.kwargs["token"]:
                return redirect(
                    'spider_base:ucontent-access',
                    token=self.object.token, access="update"
                )

            if context["form"].is_valid():
                context["form"] = self.get_form_class()(
                    **self.get_form_success_kwargs()
                )
        if isinstance(rendered, HttpResponseBase):
            return rendered

        context["content"] = rendered
        return super().render_to_response(context)

    def get_usercomponent(self):
        return get_object_or_404(
            UserComponent.objects.prefetch_related("protections"),
            contents__token=self.kwargs["token"]
        )

    def get_object(self, queryset=None):
        # can bypass idlist and searchlist with own queryset arg
        if not queryset:
            queryset = self.get_queryset()

        # required for next/previous token
        queryset = queryset.select_related(
            "usercomponent", "usercomponent__user",
            "usercomponent__user__spider_info"
        ).filter(usercomponent=self.usercomponent).order_by("priority", "id")
        ob = get_object_or_404(
            queryset,
            token=self.kwargs["token"]
        )
        ob.previous_object = prev_in_order(ob, queryset)
        ob.next_object = next_in_order(ob, queryset)
        return ob


class ContentDelete(EntityDeletionMixin, DeleteView):
    model = AssignedContent
    usercomponent = None

    def form_valid(self, form):
        _ = gettext
        messages.error(
            self.request, _('Content deleted.')
        )
        return super().form_valid(form)

    def dispatch(self, request, *args, **kwargs):
        self.object = self.get_object()
        self.usercomponent = self.get_usercomponent()
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs["uc"] = self.usercomponent
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return reverse(
            "spider_base:ucontent-list", kwargs={
                "token": self.usercomponent.token,
            }
        )

    def get_required_timedelta(self):
        _time = self.object.content.deletion_period
        if _time:
            _time = timedelta(seconds=_time)
        else:
            _time = timedelta(seconds=0)
        return _time

    def get_usercomponent(self):
        return self.object.usercomponent

    def get_object(self, queryset=None):
        if not queryset:
            queryset = self.get_queryset()
        return get_object_or_404(
            queryset, token=self.kwargs["token"]
        )



__all__ = [
    "TagLayoutForm", "TagLayoutAdminForm", "SpiderTagForm", "generate_form",
]

import json

import posixpath
from collections import OrderedDict
# from django.utils.translation import gettext_lazy as _
from django.conf import settings
from django import forms

# from django.apps import apps
from django.db.models import Q, QuerySet
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import NON_FIELD_ERRORS
from django.utils.translation import gettext_lazy as _

from rdflib import XSD

from .fields import generate_fields
from .models import TagLayout, SpiderTag

from spkcspider.apps.spider.fields import MultipleOpenChoiceField, JsonField
from spkcspider.apps.spider.widgets import ListWidget, OpenChoiceWidget
from spkcspider.apps.spider.helpers import merge_get_url
from spkcspider.apps.spider.models import (
    AssignedContent, ReferrerObject
)
from spkcspider.apps.spider.contents import BaseContent


from .widgets import SchemeWidget
from .fields import installed_fields

# don't spam set objects
_empty_set = frozenset()


_extra = '' if settings.DEBUG else '.min'


class TagLayoutForm(forms.ModelForm):
    layout = JsonField(
        widget=SchemeWidget(
            attrs={
                "field_types": json.dumps(list(installed_fields.keys()))
            }
        )
    )
    default_verifiers = MultipleOpenChoiceField(
        widget=ListWidget(
            format_type="url", item_label=_("Url to Verifier")
        ), required=False
    )

    class Meta:
        model = TagLayout
        fields = ["name", "unique", "layout", "default_verifiers"]

    usertag = None

    def _save_m2m(self):
        self.instance.usertag = self.usertag
        self.instance.save()
        self.instance.usertag.refresh_from_db()
        return super()._save_m2m()

    def clean_default_verifiers(self):
        values = self.cleaned_data["default_verifiers"]
        if not isinstance(values, list):
            raise forms.ValidationError(
                _("Invalid format")
            )
        return values

    def save(self, commit=True):
        self.usertag = self.instance.usertag
        if commit:
            self.usertag.save()
            self._save_m2m()
        else:
            self.save_m2m = self._save_m2m
        return self.usertag


class TagLayoutAdminForm(forms.ModelForm):
    layout = JsonField(
        widget=SchemeWidget(
            attrs={
                "field_types": json.dumps(list(installed_fields.keys()))
            }
        )
    )
    default_verifiers = MultipleOpenChoiceField(
        widget=ListWidget(
            format_type="url", item_label=_("Url to Verifier")
        ), required=False
    )

    class Meta:
        model = TagLayout
        fields = ["name", "unique", "layout", "default_verifiers", "usertag"]

    class Media:
        css = {
            'all': [
                'node_modules/@fortawesome/fontawesome-free/css/all.min.css'
            ]
        }

    def clean_default_verifiers(self):
        values = self.cleaned_data["default_verifiers"]
        if not isinstance(values, list):
            raise forms.ValidationError(
                _("Invalid format")
            )
        return values


class SpiderTagForm(forms.ModelForm):
    updateable_by = MultipleOpenChoiceField(
        required=False, initial=False,
        widget=ListWidget(
            format_type="url"
        )
    )
    layout = forms.ModelChoiceField(
        queryset=TagLayout.objects.none(),
        to_field_name="name"
    )

    class Meta:
        model = SpiderTag
        fields = ["layout", "updateable_by"]

    def __init__(self, user=None, **kwargs):
        super().__init__(**kwargs)
        index = user.usercomponent_set.get(name="index")
        self.fields["layout"].queryset = TagLayout.objects.filter(
            Q(usertag__isnull=True) |
            Q(usertag__associated_rel__usercomponent=index)
        ).order_by("name")

    def clean_updateable_by(self):
        values = self.cleaned_data["updateable_by"]
        values = set(map(merge_get_url, values))
        existing = ReferrerObject.objects.filter(
            url__in=values
        ).values_list("url", flat=True)
        for url in values.difference(existing):
            # extra validation
            ReferrerObject.objects.get_or_create(
                url=url
            )
        return ReferrerObject.objects.filter(
            url__in=values
        )


def generate_form(name, layout):
    _gen_fields = generate_fields(layout, "tag")
    _temp_field = forms.BooleanField(required=False, initial=False)
    setattr(_temp_field, "hashable", True)
    _gen_fields.insert(0, (
        "primary",
        _temp_field
    ))
    _temp_field = MultipleOpenChoiceField(
        required=False, initial=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 300px; width:100%"
            }
        )
    )
    setattr(_temp_field, "spkc_datatype", XSD.anyURI)

    _gen_fields.append(("updateable_by", _temp_field))
    _temp_field = MultipleOpenChoiceField(
        required=False, initial=False,
        widget=ListWidget(
            format_type="url", item_label=_("Url to Verifier")
        )
    )
    setattr(_temp_field, "spkc_datatype", XSD.anyURI)
    _gen_fields.append(("verified_by", _temp_field))

    class _form(forms.BaseForm):
        __name__ = name
        declared_fields = OrderedDict(_gen_fields)
        base_fields = declared_fields
        # used in models
        layout_generating_form = True

        class Meta:
            error_messages = {
                NON_FIELD_ERRORS: {
                    'unique_together': _(
                        'Primary layout for "%s" exists already'
                    ) % name
                }
            }

        def __init__(self, instance, *, uc=None, initial=None, **kwargs):
            if not initial:
                initial = {}
            self.instance = instance
            _initial = self.encode_initial(initial)
            _initial["primary"] = getattr(instance, "primary", False)
            _initial["verified_by"] = getattr(instance, "verified_by", [])
            super().__init__(
                initial=_initial, **kwargs
            )

            for field in self.fields.values():
                if hasattr(field, "queryset"):
                    filters = {}
                    q_user = Q()
                    q_uc = Q()
                    # can also contain __lte or so
                    attr = getattr(field, "filter_strength_link", None)
                    if attr:
                        filters[attr] = uc.strength
                    attrs = getattr(field, "filters_user", _empty_set)
                    if attrs:
                        for i in attrs:
                            q_user |= Q(**{i: uc.user})
                    attrs = getattr(field, "filters_usercomponent", _empty_set)
                    if attrs:
                        for i in attrs:
                            q_uc |= Q(**{i: uc})
                    field.queryset = field.queryset.filter(
                        q_user, q_uc, **filters
                    )

        def clean_updateable_by(self):
            values = self.cleaned_data["updateable_by"]
            values = set(map(merge_get_url, values))
            existing = ReferrerObject.objects.filter(
                url__in=values
            ).values_list("url", flat=True)
            for url in values.difference(existing):
                # extra validation
                ReferrerObject.objects.get_or_create(
                    url=url
                )
            return ReferrerObject.objects.filter(
                url__in=values
            )

        def clean(self):
            super().clean()

            _cached_references = []
            for key, value in self.cleaned_data.items():
                if key in ("verified_by", "updateable_by", "primary"):
                    continue
                # e.g. anchors
                if isinstance(value, AssignedContent):
                    _cached_references.append(value)

                if issubclass(type(value), BaseContent):
                    _cached_references.append(value.associated)

                # e.g. anchors
                if isinstance(value, QuerySet):
                    if issubclass(value.model, AssignedContent):
                        _cached_references += list(value)

                    if issubclass(value.model, BaseContent):
                        _cached_references += list(
                            AssignedContent.objects.filter(
                                object_id__in=value.values_list(
                                    "id", flat=True
                                ),
                                content_type=ContentType.objects.get_for_model(
                                    value.model
                                )
                            )
                        )
            self.instance._cached_references = _cached_references
            self.instance.full_clean()
            return self.cleaned_data

        @classmethod
        def encode_initial(cls, initial, prefix="tag", base=None):
            if base is None:
                base = {}
            for i in initial.items():
                if isinstance(i[1], dict):
                    new_prefix = posixpath.join(prefix, i[0])
                    cls.encode_initial(i[1], prefix=new_prefix, base=base)
                else:
                    base[posixpath.join(prefix, i[0])] = i[1]
            return base

        def encode_data(self, cleaned_data, prefix="tag"):
            ret = {}
            for counter, i in enumerate(cleaned_data.items()):
                if not i[0].startswith(prefix):  # unrelated data
                    continue
                selected_dict = ret
                splitted = i[0].split("/")
                # last key is item key, first is "tag"
                for key in splitted[1:-1]:
                    if key not in selected_dict:
                        selected_dict[key] = {}
                    selected_dict = selected_dict[key]
                if hasattr(self.fields[i[0]], "tagdata_from_value"):
                    selected_dict[splitted[-1]] = \
                        self.fields[i[0]].tagdata_from_value(i[1])
                else:
                    selected_dict[splitted[-1]] = i[1]
            return ret

        def save_m2m(self):
            pass

        def save(self, commit=True):
            self.instance.primary = self.cleaned_data["primary"]
            # self.instance.verified_by = self.cleaned_data["verified_by"]
            # self.instance.updateable_by = self.cleaned_data["updateable_by"]
            self.instance.tagdata = self.encode_data(self.cleaned_data)
            if commit:
                self.instance.save()

            return self.instance

    return _form



from django.contrib import admin
from django.conf import settings

from .models import (
    Protection, AssignedProtection, UserComponent, AssignedContent,
    ContentVariant, UserInfo
)


@admin.register(AssignedContent)
class AssignedContentAdmin(admin.ModelAdmin):
    fields = ['info', 'created', 'modified', 'deletion_requested', 'token']
    readonly_fields = [
        'info', 'created', 'modified'
    ]

    def has_module_permission(self, request):
        return True

    def has_delete_permission(self, request, obj=None):
        n = request.user._meta.app_label
        m = request.user._meta.model_name
        if not request.user.is_active:
            return False
        # not obj allows deletion of users
        if not obj or request.user.is_superuser:
            return True
        # obj is UserComponent
        if obj.usercomponent.name == "index":
            return request.user.has_perm("{}.delete_{}".format(n, m))
        return request.user.has_perm("spider_base.delete_assignedcontent")

    def has_view_permission(self, request, obj=None):
        if request.user.is_superuser:
            return True
        # obj is UserComponent
        if not obj or obj.usercomponent.name == "index":
            return False
        return request.user.has_perm("spider_base.view_assignedcontent")

    def has_change_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        # obj is UserComponent
        if not obj or obj.usercomponent.name == "index":
            return False
        if request.user.is_superuser:
            return True
        return request.user.has_perm("spider_base.change_assignedcontent")

    def has_add_permission(self, request, obj=None):
        return False


class AssignedProtectionInline(admin.TabularInline):
    model = AssignedProtection
    # protection field is not possible (and shouldn't, can corrupt data)
    # data is too dangerous, admin should not have access to e.g. pws
    fields = [
        'created', 'modified', 'active', 'instant_fail'
    ]
    readonly_fields = ['created', 'modified']
    fk_name = 'usercomponent'
    extra = 0

    def has_add_permission(self, request, obj=None):
        # adding protection doesn't make sense without data
        return False


class ContentInline(admin.TabularInline):
    model = AssignedContent
    fields = [
        'info', 'created', 'modified', 'deletion_requested', 'token'
    ]
    readonly_fields = [
        'info', 'created', 'modified'
    ]
    # until fixed:
    readonly_fields += ['deletion_requested', 'token']
    show_change_link = True
    extra = 0
    # content is not visible

    def has_delete_permission(self, request, obj=None):
        n = request.user._meta.app_label
        m = request.user._meta.model_name
        if not request.user.is_active:
            return False
        # not obj allows deletion of users
        if not obj or request.user.is_superuser:
            return True
        # obj is UserComponent
        if obj.name == "index":
            return request.user.has_perm("{}.delete_{}".format(n, m))
        return request.user.has_perm("spider_base.delete_usercontent")

    def has_view_permission(self, request, obj=None):
        if request.user.is_superuser:
            return True
        # obj is UserComponent
        if not obj or obj.name == "index":
            return False
        return request.user.has_perm("spider_base.view_usercontent")

    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False


@admin.register(UserComponent)
class UserComponentAdmin(admin.ModelAdmin):
    inlines = [
        ContentInline,
        AssignedProtectionInline,
    ]
    actions = ["feature", "unfeature"]
    fields = [
        'user', 'name', 'created', 'modified', 'description', 'features',
        'can_auth', 'allow_domain_mode',
        'featured', 'public', 'strength', 'token', 'required_passes',
        'deletion_requested', 'token_duration'
    ]
    readonly_fields = [
        'created', 'modified', 'featured', 'strength',
        'can_auth', 'allow_domain_mode'
    ]
    list_display = ('name', 'username', 'strength', 'featured', 'modified')
    view_on_site = True

    def has_module_permission(self, request):
        return True

    def feature(self, request, queryset):
        queryset.exclude(
            name__in=("index", "fake_index")
        ).exclude(public=False).update(featured=True)
    feature.allowed_permissions = ('can_feature',)

    def unfeature(self, request, queryset):
        queryset.update(featured=False)
    unfeature.allowed_permissions = ('can_feature',)

    def has_can_feature_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        if request.user.is_superuser:
            return True
        return request.user.has_perm("spider_base.can_feature")

    def has_add_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        if request.user.is_superuser:
            return True
        return request.user.has_perm("spider_base.add_usercomponent")

    def has_delete_permission(self, request, obj=None):
        n = request.user._meta.app_label
        m = request.user._meta.model_name
        if not request.user.is_active:
            return False
        if request.user.is_superuser:
            return True
        # you must be able to delete user to delete index component
        if obj and obj.name == "index":
            if not request.user.has_perm("{}.delete_{}".format(n, m)):
                return False
        return request.user.has_perm("spider_base.delete_usercomponent")

    def has_view_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        if request.user.is_superuser:
            return True
        if obj and obj.name == "index" and obj.user != request.user:
            return False
        return request.user.has_perm("spider_base.view_usercomponent")

    def has_change_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        if request.user.is_superuser:
            return True
        if obj and obj.name == "index" and obj.user != request.user:
            return False
        return request.user.has_perm("spider_base.change_usercomponent")


@admin.register(Protection)
class ProtectionAdmin(admin.ModelAdmin):
    view_on_site = False
    fields = ['code']

    def has_module_permission(self, request):
        return True

    def has_view_permission(self, request, obj=None):
        return True

    def has_change_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        return request.user.is_superuser or request.user.is_staff

    def has_add_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        return request.user.is_superuser or request.user.is_staff

    def has_delete_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        return request.user.is_superuser or request.user.is_staff


@admin.register(ContentVariant)
class ContentVariantAdmin(ProtectionAdmin):
    fields = ['name']


@admin.register(UserInfo)
class UserInfoAdmin(admin.ModelAdmin):
    view_on_site = False
    fields = []

    def get_actions(self, request):
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions

    def has_view_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        return request.user.is_superuser or request.user.is_staff

    def has_change_permission(self, request, obj=None):
        if not request.user.is_active:
            return False
        # not obj allows deletion of users
        if not obj:
            return True
        return False

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        n = request.user._meta.app_label
        m = request.user._meta.model_name
        if not request.user.is_active:
            return False
        # not obj allows deletion of users
        if not obj:
            return True
        return request.user.has_perm("{}.delete_{}".format(n, m))


if 'django.contrib.flatpages' in settings.INSTALLED_APPS:
    from django.contrib.flatpages.admin import FlatPageAdmin
    from .forms.flatpages import FlatpageItemForm

    FlatPageAdmin.form = FlatpageItemForm



"""
installed_protections, add_protection
namespace: spider_base

"""

__all__ = ("installed_protections", "BaseProtection", "ProtectionResult",
           "initialize_protection_models")

from random import SystemRandom

from django.conf import settings
from django import forms
from django.http import Http404
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext
from django.contrib.auth import authenticate
from django.views.decorators.debug import sensitive_variables
# from django.contrib.auth.hashers import make_password

from .helpers import add_by_field
from django.utils.crypto import constant_time_compare
from .constants import ProtectionType, ProtectionResult, index_names
from .fields import MultipleOpenChoiceField
from .widgets import OpenChoiceWidget

installed_protections = {}
_sysrand = SystemRandom()

# don't spam set objects
_empty_set = frozenset()

# for debug/min switch
_extra = '' if settings.DEBUG else '.min'


def initialize_protection_models(apps=None):
    if not apps:
        from django.apps import apps
    UserComponent = apps.get_model("spider_base", "UserComponent")
    AssignedProtection = apps.get_model("spider_base", "AssignedProtection")
    ProtectionModel = apps.get_model("spider_base", "Protection")
    for code, val in installed_protections.items():
        ret = ProtectionModel.objects.get_or_create(
            defaults={"ptype": val.ptype}, code=code
        )[0]
        if ret.ptype != val.ptype:
            ret.ptype = val.ptype
            ret.save()

    login = ProtectionModel.objects.filter(code="login").first()
    if login:
        for uc in UserComponent.objects.filter(name="index"):
            asuc = AssignedProtection.objects.get_or_create(
                defaults={"active": True},
                usercomponent=uc, protection=login
            )[0]
            if not asuc.active:
                asuc.active = True
                asuc.save()

    UserComponent.objects.filter(name__in=index_names).update(strength=10)

    invalid_models = ProtectionModel.objects.exclude(
        code__in=installed_protections.keys()
    )
    if invalid_models.exists():
        print("Invalid protections, please update or remove them:",
              [t.code for t in invalid_models])


class BaseProtection(forms.Form):
    """
        Base for Protections
        Usage: use form to define some configurable fields for configuration
        use auth to validate: in this case:
            template_name, render, form variable are used
    """
    use_required_attribute = False
    active = forms.BooleanField(label=_("Active"), required=False)
    instant_fail = forms.BooleanField(
        label=_("Instant fail"), required=False,
        help_text=_("Fail instantly if not fullfilled. "
                    "Don't count to required_passes.")
    )
    # unique code name max 10 slug chars
    # if imported by extract_app_dicts, name is automatically set to key name
    # name = foo

    # ptype valid for, is overwritten with current ptype
    ptype = ProtectionType.access_control.value

    # description of Protection
    description = None

    template_name = None
    # form for authentication
    form = None
    # optional render function
    # render = None

    @classmethod
    def render_raw(cls, result):
        return {}

    # auto populated, instance
    protection = None
    instance = None

    # initial values
    initial = {}

    def __init__(self, protection, ptype, request, assigned=None, **kwargs):
        self.protection = protection
        self.ptype = ptype
        if assigned:
            self.instance = assigned.filter(protection=self.protection).first()
        initial = self.get_initial()
        # does assigned protection exist?
        if self.instance:
            # if yes force instance.active information
            initial["active"] = self.instance.active
            initial["instant_fail"] = self.instance.instant_fail
        super().__init__(initial=initial, **kwargs)
        self.fields["active"].help_text = self.description

    def get_initial(self):
        initial = self.initial.copy()
        if self.instance:
            initial.update(self.instance.data)
        return initial

    def get_strength(self):
        # can provide strength in range 0-4
        # 0 no protection
        # 1 weak protection
        # 2 normal protection
        # 3 strong protection
        # 4 reserved for login only, can be returned to auth user
        # tuple for min max
        return (1, 1)

    @staticmethod
    def extract_form_kwargs(request):
        kwargs = {}
        if request.method in ["POST", "PUT"]:
            kwargs = {
                'data': request.POST,
                'files': request.FILES,
            }
        return kwargs

    @classmethod
    def auth(cls, request, **kwargs):
        if hasattr(cls, "auth_form"):
            form = cls.auth_form(**cls.extract_form_kwargs(request))
            if form.is_valid():
                return 1
            return form
        return False

    @classmethod
    def localize_name(cls, name=None):
        if not name:
            name = cls.name
        return pgettext("protection name", name.title())

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name(name)

    def __str__(self):
        return self.localize_name(self.name)


# only friends have access
@add_by_field(installed_protections, "name")
class FriendProtection(BaseProtection):
    name = "friends"
    ptype = ProtectionType.access_control.value

    users = MultipleOpenChoiceField(
        label=_("Users"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )
    description = _("Limit access to selected users")

    def get_strength(self):
        return (3, 3)

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if (
            obj and
            (
                getattr(request.user, request.user.USERNAME_FIELD) in
                obj.data["users"]
            )
        ):
            return 3
        else:
            return False


@add_by_field(installed_protections, "name")
class RandomFailProtection(BaseProtection):
    name = "randomfail"
    ptype = ProtectionType.access_control.value
    ptype += ProtectionType.authentication.value
    ptype += ProtectionType.side_effects.value

    success_rate = forms.IntegerField(
        label=_("Success Rate"), min_value=20, max_value=100, initial=100,
        widget=forms.NumberInput(attrs={'type': 'range'}),
        help_text=_("Set success rate")
    )

    use_404 = forms.BooleanField(label="Use 404 errors?", required=False)

    description = _(
        "Fail/Refuse randomly. Optionally with 404 error, "
        "to disguise correct access."
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if ProtectionType.authentication.value in self.ptype:
            # login should not be possible alone with it
            self.fields["instant_fail"].initial = True
            self.initial["instant_fail"] = True
            self.fields["instant_fail"].disabled = True

    def get_strength(self):
        if self.cleaned_data["success_rate"] > 70:
            return (0, 0)
        return (1, 1)

    @classmethod
    def localize_name(cls, name=None):
        return pgettext("protection name", "Random Fail")

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if obj and obj.data.get("success_rate", None):
            if _sysrand.randrange(1, 101) <= obj.data["success_rate"]:
                return 0
            elif obj.data.get("use_404", False):
                raise Http404()
        return False


@add_by_field(installed_protections, "name")
class LoginProtection(BaseProtection):
    name = "login"
    ptype = ProtectionType.authentication.value
    ptype += ProtectionType.access_control.value

    description = _("Use Login password")

    allow_auth = forms.BooleanField(
        label=_("Component authentication"), required=False
    )

    class auth_form(forms.Form):
        use_required_attribute = False
        password = forms.CharField(
            label=_("Password"),
            strip=False,
            widget=forms.PasswordInput,
        )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if ProtectionType.authentication.value in self.ptype:
            self.fields["active"].initial = True
            self.initial["active"] = True
            self.fields["active"].disabled = True
            del self.fields["allow_auth"]

    def get_strength(self):
        return (3, 4 if self.cleaned_data.get("allow_auth", False) else 3)

    @classmethod
    @sensitive_variables("password")
    def auth(cls, request, obj, **kwargs):
        if not obj:
            return cls.auth_form()

        username = obj.usercomponent.username
        for password in request.POST.getlist("password")[:2]:
            if authenticate(
                request, username=username, password=password, nospider=True
            ):
                if obj.data.get("allow_auth", False):
                    return 4
                return 3
        return cls.auth_form()

    @classmethod
    def auth_localize_name(cls, name=None):
        return cls.localize_name("Password")


@add_by_field(installed_protections, "name")
class PasswordProtection(BaseProtection):
    name = "password"
    ptype = ProtectionType.access_control.value
    ptype += ProtectionType.authentication.value

    description = _("Protect with extra passwords")
    prefix = "protection_passwords"

    class auth_form(forms.Form):
        use_required_attribute = False
        password = forms.CharField(
            label=_("Extra Password"),
            strip=False,
            widget=forms.PasswordInput,
        )

    passwords = MultipleOpenChoiceField(
        label=_("Passwords"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    auth_passwords = MultipleOpenChoiceField(
        label=_("Passwords (for component authentication)"), required=False,
        widget=OpenChoiceWidget(
            allow_multiple_selected=True,
            attrs={
                "style": "min-width: 250px; width:100%"
            }
        )
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if ProtectionType.authentication.value in self.ptype:
            del self.fields["auth_passwords"]

    @staticmethod
    def eval_strength(length):
        if not length:
            return 0
        elif length > 15:
            return 2
        elif length > 40:
            return 3
        return 1

    def get_strength(self):
        maxstrength = self.eval_strength(self.cleaned_data["max_length"])
        if len(self.cleaned_data["auth_passwords"]) > 0:
            maxstrength = 4
        return (
            self.eval_strength(self.cleaned_data["min_length"]),
            maxstrength
        )

    def clean_passwords(self):
        passwords = set()
        for pw in self.cleaned_data["passwords"]:
            if len(pw) > 0:
                passwords.add(pw)
        return list(passwords)

    def clean_auth_passwords(self):
        passwords = set()
        for pw in self.cleaned_data["auth_passwords"]:
            if self.eval_strength(len(pw)) >= 2:
                passwords.add(pw)
        return list(passwords)

    def clean(self):
        ret = super().clean()
        # prevents user self lockout
        if ProtectionType.authentication.value in self.ptype and \
           len(self.cleaned_data["passwords"]) == 0:
            self.cleaned_data["active"] = False

        min_length = None
        max_length = None

        for pw in self.cleaned_data.get("apasswords", _empty_set):
            lenpw = len(pw)
            if not min_length or lenpw < min_length:
                min_length = lenpw
            if not max_length or lenpw > max_length:
                max_length = lenpw

        for pw in self.cleaned_data.get("auth_passwords", _empty_set):
            lenpw = len(pw)
            if not min_length or lenpw < min_length:
                min_length = lenpw
            if not max_length or lenpw > max_length:
                max_length = lenpw
        ret["min_length"] = min_length
        ret["max_length"] = max_length
        return ret

    @classmethod
    @sensitive_variables("password", "pw")
    def auth(cls, request, obj, **kwargs):
        retfalse = cls.auth_form()
        if not obj:
            retfalse.fields["password"].label = _(
                "Extra Password (if required)"
            )
            return retfalse
        success = False
        auth = False
        max_length = 0
        for password in request.POST.getlist("password")[:2]:
            for pw in obj.data["passwords"]:
                if constant_time_compare(pw, password):
                    success = True
            if success:
                max_length = max(len(password), max_length)

        for password in request.POST.getlist("password")[:2]:
            for pw in obj.data["auth_passwords"]:
                if constant_time_compare(pw, password):
                    success = True
            if success:
                max_length = max(len(password), max_length)
                auth = True

        if success:
            if auth:
                return 4
            return cls.eval_strength(max_length)
        return retfalse


if getattr(settings, "USE_CAPTCHAS", False):
    from captcha.fields import CaptchaField

    @add_by_field(installed_protections, "name")
    class CaptchaProtection(BaseProtection):
        name = "captcha"
        ptype = ProtectionType.access_control.value
        ptype += ProtectionType.authentication.value
        description = _("Require captcha")

        class auth_form(forms.Form):
            use_required_attribute = False
            prefix = "protection_captcha"

        auth_form.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
            CaptchaField(label=_("Captcha"))
        auth_form.base_fields[settings.SPIDER_CAPTCHA_FIELD_NAME] = \
            auth_form.declared_fields[settings.SPIDER_CAPTCHA_FIELD_NAME]

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            if ProtectionType.authentication.value in self.ptype:
                # login should not be possible with captcha alone
                self.initial["instant_fail"] = True
                self.fields["instant_fail"].initial = True
                self.fields["instant_fail"].disabled = True
                self.fields["instant_fail"].help_text = \
                    _("instant_fail is for login required")

                if getattr(settings, "REQUIRE_LOGIN_CAPTCHA", False):
                    self.initial["active"] = True
                    self.fields["active"].disabled = True
                    self.fields["instant_fail"].help_text = \
                        _("captcha is for login required (admin setting)")

        def get_strength(self):
            return (1, 1)

        @classmethod
        def localize_name(cls, name=None):
            return pgettext("protection name", "Captcha Protection")

        @classmethod
        def auth(cls, request, obj, **kwargs):
            if not obj:
                return cls.auth_form()
            form = cls.auth_form(**cls.extract_form_kwargs(request))
            if request.method != "GET" and form.is_valid():
                return 1
            return form


# travel protection
@add_by_field(installed_protections, "name")
class TravelProtection(BaseProtection):
    name = "travel"
    ptype = ProtectionType.access_control.value
    ptype += ProtectionType.authentication.value

    description = _("Deny access if valid travel protection is active")

    def get_strength(self):
        return (0, 0)

    @classmethod
    def auth(cls, request, obj, **kwargs):
        if obj:
            from .models import TravelProtection as TravelProtectionContent
            travel = TravelProtectionContent.objects.get_active()
            # simple: disallow cannot be changed by user in fake mode
            travel = travel.filter(
                disallow__contains=obj.usercomponent
            )
            if not travel.exists():
                return 0
        return False



# Generated by Django 2.1.5 on 2019-01-18 10:17

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('spider_base', '0011_auto_20190117_1256'),
    ]

    operations = [
        migrations.AlterField(
            model_name='linkcontent',
            name='content',
            field=models.ForeignKey(limit_choices_to={'strength_link__lte': 10}, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='primary_anchor',
            field=models.ForeignKey(blank=True, help_text='Select main identyifing anchor. Also used for attaching persisting tokens (elsewise they are attached to component)', limit_choices_to={'info__contains': '\nanchor\n'}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='primary_anchor_for', to='spider_base.AssignedContent'),
        ),
        migrations.AlterField(
            model_name='usercomponent',
            name='public',
            field=models.BooleanField(default=False, help_text='Is public? Is listed and searchable?<br/>Note: This field is maybe not deactivatable because of assigned content'),
        ),
    ]



# Generated by Django 2.2 on 2019-04-03 04:17

from django.db import migrations, models
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('spider_filets', '0009_textfilet_push'),
    ]

    operations = [
        migrations.AddField(
            model_name='filefilet',
            name='license',
            field=models.TextField(default='', blank=True),
        ),
        migrations.AddField(
            model_name='filefilet',
            name='license_name',
            field=models.CharField(default='other', max_length=255),
        ),
        migrations.AddField(
            model_name='filefilet',
            name='sources',
            field=jsonfield.fields.JSONField(default=list, blank=True),
        ),
        migrations.AddField(
            model_name='textfilet',
            name='license',
            field=models.TextField(default='', blank=True),
        ),
        migrations.AddField(
            model_name='textfilet',
            name='license_name',
            field=models.CharField(default='other', max_length=255),
        ),
        migrations.AddField(
            model_name='textfilet',
            name='sources',
            field=jsonfield.fields.JSONField(default=list, blank=True),
        ),
    ]
